from __future__ import division

import numpy as np
import itertools

from ..misc import merge
from .euclidean import EuclideanDomain
from .domain import Domain
from .domain import TOL, DEFAULT_CVXPY_KWARGS

class TensorProductDomain(EuclideanDomain):
	r""" A class describing a tensor product of a multiple domains


	Parameters
	----------
	domains: list of domains
		Domains to combine into a single domain
	**kwargs
		Additional keyword arguments to pass to CVXPY
	"""
	def __init__(self, domains = None, **kwargs):
		self._domains = []
		if domains == None:
			domains = []
	
		for domain in domains:
			assert isinstance(domain, Domain), "Input must be list of domains"
			if isinstance(domain, TensorProductDomain):
				# If one of the sub-domains is a tensor product domain,
				# flatten it to limit recursion
				self._domains.extend(domain.domains)
			else:
				self._domains.append(domain)
		self.kwargs = merge(DEFAULT_CVXPY_KWARGS, kwargs)

	def __str__(self):
		return "<TensorProductDomain on R^%d of %d domains>" % (len(self), len(self.domains))
		
	@property
	def names(self):
		return list(itertools.chain(*[dom.names for dom in self.domains]))

	def _is_linquad_domain(self):
		return all([dom.is_linquad_domain for dom in self.domains])
	
	def _is_linineq_domain(self):
		return all([dom.is_linineq_domain for dom in self.domains])
		
	def _is_box_domain(self):
		return all([dom.is_box_domain for dom in self.domains])

	@property
	def domains(self):
		return self._domains

	@property
	def _slices(self):	
		start, stop = 0,0
		for dom in self.domains:
			stop += len(dom)
			yield(slice(start, stop))
			start += len(dom) 	

	def _sample(self, draw = 1):
		X = []
		for dom in self.domains:
			X.append(dom.sample(draw = draw))
		return np.hstack(X)

	def _isinside(self, X, tol = TOL):
		inside = np.ones(X.shape[0], dtype = np.bool)
		for dom, I in zip(self.domains, self._slices):
			#print(dom, I, dom.isinside(X[:,I]))
			inside = inside & dom.isinside(X[:,I], tol = tol)
		return inside

	def _extent(self, x, p):
		alpha = [dom._extent(x[I], p[I]) for dom, I in zip(self.domains, self._slices)]
		return min(alpha)


	def __len__(self):
		return sum([len(dom) for dom in self.domains])


	################################################################################		
	# Normalization related functions
	################################################################################		
	
	def _normalize(self, X):
		return np.hstack([dom.normalize(X[:,I]) for dom, I in zip(self.domains, self._slices)])

	def _unnormalize(self, X_norm):
		return np.hstack([dom.unnormalize(X_norm[:,I]) for dom, I in zip(self.domains, self._slices)])

	def _unnormalize_der(self):
		return np.diag(np.hstack([np.diag(dom._unnormalize_der()) for dom in self.domains]))

	def _normalize_der(self):
		return np.diag(np.hstack([np.diag(dom._normalize_der()) for dom in self.domains]))

	def _normalized_domain(self, **kwargs):
		domains_norm = [dom.normalized_domain(**kwargs) for dom in self.domains]
		return TensorProductDomain(domains = domains_norm)

	
	################################################################################		
	# Convex solver problems
	################################################################################		
	
	def _build_constraints_norm(self, x_norm):
		constraints = []
		for dom, I in zip(self.domains, self._slices):
			constraints.extend(dom._build_constraints_norm(x_norm[I]))
		return constraints 


	def _build_constraints(self, x):
		constraints = []
		for dom, I in zip(self.domains, self._slices):
			constraints.extend(dom._build_constraints(x[I]))
		return constraints 

	################################################################################		
	# Properties resembling LinQuad Domains 
	################################################################################		

	@property
	def lb(self):
		return np.concatenate([dom.lb for dom in self.domains])

	@property
	def ub(self):
		return np.concatenate([dom.ub for dom in self.domains])

	@property
	def A(self):
		A = []
		for dom, I in zip(self.domains, self._slices):
			A_tmp = np.zeros((dom.A.shape[0] ,len(self)))
			A_tmp[:,I] = dom.A
			A.append(A_tmp)
		return np.vstack(A)

	@property
	def b(self):
		return np.concatenate([dom.b for dom in self.domains])

	@property
	def A_eq(self):
		A_eq = []
		for dom, I in zip(self.domains, self._slices):
			A_tmp = np.zeros((dom.A_eq.shape[0] ,len(self)))
			A_tmp[:,I] = dom.A_eq
			A_eq.append(A_tmp)
		return np.vstack(A_eq)
	
	@property
	def b_eq(self):
		return np.concatenate([dom.b_eq for dom in self.domains])

	def add_constraints(self, *args, **kwargs):
		if self.is_linquad_domain:
			from .linquad import LinQuadDomain
			return LinQuadDomain.add_constraints(self, *args, **kwargs)
		else:
			raise NotImplementedError

	def _init_lb(self, *args, **kwargs):
		from .linquad import LinQuadDomain
		return LinQuadDomain._init_lb(self, *args, **kwargs)
	
	def _init_ub(self, *args, **kwargs):
		from .linquad import LinQuadDomain
		return LinQuadDomain._init_ub(self, *args, **kwargs)
	
	def _init_ineq(self, *args, **kwargs):
		from .linquad import LinQuadDomain
		return LinQuadDomain._init_ineq(self, *args, **kwargs)
	
	def _init_eq(self, *args, **kwargs):
		from .linquad import LinQuadDomain
		return LinQuadDomain._init_eq(self, *args, **kwargs)

	def _init_quad(self, *args, **kwargs):
		from .linquad import LinQuadDomain
		return LinQuadDomain._init_quad(self, *args, **kwargs)




import numpy as np
import pandas as pd
from scipy.stats import norm
import unittest

from context import grama as gr
from context import models
from context import data

X = gr.Intention()

## Core function tests
##################################################
class TestFits(unittest.TestCase):
    """Test implementations of fitting procedures"""

    def setUp(self):
        ## Smooth model
        self.md_smooth = (
            gr.Model()
            >> gr.cp_function(fun=lambda x: [x, x + 1], var=["x"], out=["y", "z"])
            # >> gr.cp_vec_function(fun=lambda df: gr.df_make(y=df.x, z=df.x + 1), var=["x"], out=["y", "z"])
            >> gr.cp_marginals(x={"dist": "uniform", "loc": 0, "scale": 2})
            >> gr.cp_copula_independence()
        )
        self.df_smooth = self.md_smooth >> gr.ev_df(df=pd.DataFrame(dict(x=[0, 1, 2])))

        ## Tree model
        self.md_tree = (
            gr.Model()
            # >> gr.cp_function(fun=lambda x: [0, x < 5], var=["x"], out=["y", "z"])
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(y=df.x, z=df.x + 1), var=["x"], out=["y", "z"]
            )
            >> gr.cp_marginals(x={"dist": "uniform", "loc": 0, "scale": 2})
            >> gr.cp_copula_independence()
        )

        self.df_tree = self.md_tree >> gr.ev_df(
            df=pd.DataFrame(dict(x=np.linspace(0, 10, num=8)))
        )

        ## Cluster model
        self.df_cluster = pd.DataFrame(
            dict(
                x=[0.1, 0.2, 0.3, 0.4, 1.1, 1.2, 1.3, 1.4],
                y=[0.3, 0.2, 0.1, 0.0, 1.3, 1.2, 1.1, 1.0],
                c=[0, 0, 0, 0, 1, 1, 1, 1],
            )
        )

    def test_gp(self):
        ## Fit routine creates usable model
        md_fit = gr.fit_gp(self.df_smooth, md=self.md_smooth)
        df_res = gr.eval_df(md_fit, self.df_smooth[self.md_smooth.var])

        ## GP provides std estimates
        self.assertTrue("y_sd" in df_res.columns)

        ## GP is an interpolation
        self.assertTrue(
            gr.df_equal(
                df_res[["x", "y_mean", "z_mean"]].rename(
                    {"y_mean": "y", "z_mean": "z"}, axis=1
                ),
                self.df_smooth,
                close=True,
            )
        )

        ## Fit copies model data
        self.assertTrue(set(md_fit.var) == set(self.md_smooth.var))
        self.assertTrue(
            set(md_fit.out)
            == set(map(lambda s: s + "_mean", self.md_smooth.out)).union(
                set(map(lambda s: s + "_sd", self.md_smooth.out))
            )
        )

    def test_rf(self):
        ## Fit routine creates usable model
        md_fit = gr.fit_rf(
            self.df_tree,
            md=self.md_tree,
            max_depth=1,  # True tree is a stump
            seed=101,
        )
        df_res = gr.eval_df(md_fit, self.df_tree[self.md_tree.var])

        ## RF can approximately recover a tree; check ends only
        self.assertTrue(
            gr.df_equal(
                df_res[["y_mean", "z_mean"]].iloc[[0, 1, -2, -1]],
                self.df_tree[["y", "z"]].iloc[[0, 1, -2, -1]]
                >> gr.tf_rename(y_mean="y", z_mean="z"),
                close=True,
                precision=1,
            )
        )

        ## Fit copies model data
        self.assertTrue(set(md_fit.var) == set(self.md_tree.var))
        self.assertTrue(
            set(md_fit.out) == set(map(lambda s: s + "_mean", self.md_tree.out))
        )

    def test_lm(self):
        ## Fit routine creates usable model
        md_fit = gr.fit_lm(
            self.df_smooth,
            md=self.md_smooth,
        )
        df_res = gr.eval_df(md_fit, self.df_smooth[self.md_smooth.var])

        ## LM can recover a linear model
        self.assertTrue(
            gr.df_equal(
                df_res,
                self.df_smooth >> gr.tf_rename(y_mean="y", z_mean="z"),
                close=True,
                precision=1,
            )
        )

        ## Fit copies model data
        self.assertTrue(set(md_fit.var) == set(self.md_smooth.var))
        self.assertTrue(
            set(md_fit.out) == set(map(lambda s: s + "_mean", self.md_smooth.out))
        )

    def test_kmeans(self):
        ## Fit routine creates usable model
        var = ["x", "y"]
        md_fit = gr.fit_kmeans(self.df_cluster, var=var, n_clusters=2)
        df_res = gr.eval_df(md_fit, self.df_cluster[var])

        ## Check correctness
        # Match clusters by min(x)
        id_true = (self.df_cluster >> gr.tf_filter(X.x == gr.min(X.x))).c[0]
        id_res = (df_res >> gr.tf_filter(X.x == gr.min(X.x))).cluster_id[0]

        df_res1 = (
            self.df_cluster >> gr.tf_filter(X.c == id_true) >> gr.tf_select(X.x, X.y)
        )
        df_res2 = (
            df_res >> gr.tf_filter(X.cluster_id == id_res) >> gr.tf_select(X.x, X.y)
        )

        self.assertTrue(gr.df_equal(df_res1, df_res2))

    def test_nls(self):
        ## Ground-truth model
        c_true = 2
        a_true = 1

        md_true = (
            gr.Model()
            >> gr.cp_function(
                fun=lambda x, epsilon: a_true * np.exp(x * c_true) + epsilon,
                var=["x", "epsilon"],
                out=["y"],
            )
            >> gr.cp_marginals(epsilon={"dist": "norm", "loc": 0, "scale": 0.5})
            >> gr.cp_copula_independence()
        )
        df_data = md_true >> gr.ev_sample(
            n=5, seed=101, df_det=gr.df_make(x=[0, 1, 2, 3, 4])
        )

        ## Model to fit
        md_param = (
            gr.Model()
            >> gr.cp_function(
                fun=lambda x, c, a: a * np.exp(x * c),
                var=["x", "c", "a"],
                out=["y"]
            )
            >> gr.cp_bounds(c=[0, 4], a=[0.1, 2.0])
        )

        ## Fit the model
        md_fit = df_data >> gr.ft_nls(
            md=md_param,
            verbose=False,
            uq_method="linpool",
        )

        ## Unidentifiable model throws warning
        # -------------------------
        md_unidet = (
            gr.Model()
            >> gr.cp_function(
                fun=lambda x, c, a, z: a / z * np.exp(x * c),
                var=["x", "c", "a", "z"],
                out=["y"],
            )
            >> gr.cp_bounds(c=[0, 4], a=[0.1, 2.0], z=[0, 1])
        )
        with self.assertWarns(RuntimeWarning):
            gr.fit_nls(
                df_data,
                md=md_unidet,
                uq_method="linpool",
            )

        ## True parameters in wide confidence region
        # -------------------------
        alpha = 1e-3
        self.assertTrue(
            (md_fit.density.marginals["c"].q(alpha / 2) <= c_true)
            and (c_true <= md_fit.density.marginals["c"].q(1 - alpha / 2))
        )

        self.assertTrue(
            (md_fit.density.marginals["a"].q(alpha / 2) <= a_true)
            and (a_true <= md_fit.density.marginals["a"].q(1 - alpha / 2))
        )

        ## Model with fixed parameter
        # -------------------------
        md_fixed = (
            gr.Model()
            >> gr.cp_function(
                fun=lambda x, c, a, e: a * np.exp(x * c) + e,
                var=["x", "c", "a", "e"],
                out=["y"],
            )
            >> gr.cp_bounds(c=[0, 4], a=[1, 1], e=[0, 0])
        )
        md_fit_fixed = df_data >> gr.ft_nls(
            md=md_fixed, verbose=False, uq_method="linpool"
        )

        # Test that fixed model can evaluate successfully
        gr.eval_sample(md_fit_fixed, n=1, df_det="nom")

        ## Model with fixed parameter and two inputs
        # -------------------------
        df_data2 = (
            gr.df_grid(x=(1,2,3,4), y=(1,2,3,4))
            >> gr.tf_mutate(f=2*X.x + X.y**2)
        )
        md_fixed2 = (
            gr.Model()
            >> gr.cp_function(
                fun=lambda x, y, a, b: a*x + b*y**2,
                var=["x", "y", "a", "b"],
                out=["f"],
            )
            >> gr.cp_bounds(a=(0, 3), b=(1,1))
        )
        md_fit_fixed2 = df_data2 >> gr.ft_nls(
            md=md_fixed2, verbose=False, uq_method="linpool"
        )

        # Test that fixed model can evaluate successfully
        gr.eval_sample(md_fit_fixed2, n=1, df_det="nom")

        ## Trajectory model
        # -------------------------
        md_base = models.make_trajectory_linear()
        md_fit = data.df_trajectory_windowed >> gr.ft_nls(
            md=md_base, method="SLSQP", tol=1e-3
        )
        df_tmp = md_fit >> gr.ev_nominal(df_det="nom")

        ## Select output for fitting
        # -------------------------
        # Split model has inconsistent "true" parameter value
        md_split = (
            gr.Model("Split")
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(
                    f=1 * df.c * df.x,
                    g=2 * df.c * df.x,
                ),
                var=["c", "x"],
                out=["f", "g"],
            )
            >> gr.cp_bounds(
                x=(-1, +1),
                c=(-1, +1),
            )
        )

        df_split = gr.df_make(x=gr.linspace(-1, +1, 100)) >> gr.tf_mutate(f=X.x, g=X.x)

        # Fitting both outputs: cannot achieve mse ~= 0
        df_both = (
            df_split
            >> gr.ft_nls(md_split, out=["f", "g"])
            >> gr.ev_df(df_split >> gr.tf_rename(f_t=X.f, g_t=X.g))
            >> gr.tf_summarize(
                mse_f=gr.mse(X.f, X.f_t),
                mse_g=gr.mse(X.g, X.g_t),
            )
        )
        self.assertTrue(df_both.mse_f[0] > 0)
        self.assertTrue(df_both.mse_g[0] > 0)

        # Fitting "f" only
        df_f = (
            df_split
            >> gr.ft_nls(md_split, out=["f"])
            >> gr.ev_df(df_split >> gr.tf_rename(f_t=X.f, g_t=X.g))
            >> gr.tf_summarize(
                mse_f=gr.mse(X.f, X.f_t),
                mse_g=gr.mse(X.g, X.g_t),
            )
        )
        self.assertTrue(df_f.mse_f[0] < 1e-16)
        self.assertTrue(df_f.mse_g[0] > 0)

        # Fitting "g" only
        df_g = (
            df_split
            >> gr.ft_nls(md_split, out=["g"])
            >> gr.ev_df(df_split >> gr.tf_rename(f_t=X.f, g_t=X.g))
            >> gr.tf_summarize(
                mse_f=gr.mse(X.f, X.f_t),
                mse_g=gr.mse(X.g, X.g_t),
            )
        )
        self.assertTrue(df_g.mse_f[0] > 0)
        self.assertTrue(df_g.mse_g[0] < 1e-16)



import numpy as np
import pandas as pd
import unittest
from scipy.stats import norm

from context import grama as gr
from context import data

X = gr.Intention()

##==============================================================================
## transform summary functions
##==============================================================================


class TestSummaryFcn(unittest.TestCase):
    def test_mean(self):
        df = data.df_diamonds >> gr.tf_select(X.cut, X.x) >> gr.tf_head(5)
        # straight summarize
        t = df >> gr.tf_summarize(m=gr.mean(X.x))
        df_truth = pd.DataFrame({"m": [4.086]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(m=gr.mean(X.x))
        df_truth = pd.DataFrame(
            {"cut": ["Good", "Ideal", "Premium"], "m": [4.195, 3.950, 4.045]}
        )
        self.assertTrue(t.equals(df_truth))
        # straight mutate
        t = df >> gr.tf_mutate(m=gr.mean(X.x))
        df_truth = df.copy()
        df_truth["m"] = df_truth.x.mean()
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(m=gr.mean(X.x))
        df_truth["m"] = pd.Series([3.950, 4.045, 4.195, 4.045, 4.195])
        self.assertTrue(t.sort_index().equals(df_truth))

    def test_skew(self):
        df_truth = pd.DataFrame({"m": [0.09984760044443139]})
        df_res = (
            data.df_shewhart
            >> gr.tf_summarize(m=gr.skew(X.tensile_strength))
        )
        self.assertTrue(df_truth.equals(df_res))

    def test_kurt(self):
        df_truth = pd.DataFrame({"m": [2.605643942300021]})
        df_res = (
            data.df_shewhart
            >> gr.tf_summarize(m=gr.kurt(X.tensile_strength))
        )
        self.assertTrue(df_truth.equals(df_res))

    def test_first(self):
        df = data.df_diamonds >> gr.tf_select(X.cut, X.x) >> gr.tf_head(5)
        # straight summarize
        t = df >> gr.tf_summarize(f=gr.first(X.x))
        df_truth = pd.DataFrame({"f": [3.95]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(f=gr.first(X.x))
        df_truth = pd.DataFrame(
            {"cut": ["Good", "Ideal", "Premium"], "f": [4.05, 3.95, 3.89]}
        )
        self.assertTrue(t.equals(df_truth))
        # summarize with order_by
        t = df >> gr.tf_summarize(f=gr.first(X.x, order_by=gr.desc(X.cut)))
        df_truth = pd.DataFrame({"f": [3.89]})
        # straight mutate
        t = df >> gr.tf_mutate(f=gr.first(X.x))
        df_truth = df.copy()
        df_truth["f"] = df_truth.x.iloc[0]
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(f=gr.first(X.x))
        df_truth["f"] = pd.Series([3.95, 3.89, 4.05, 3.89, 4.05])
        self.assertTrue(t.sort_index().equals(df_truth))

    def test_last(self):
        df = data.df_diamonds >> gr.tf_select(X.cut, X.x) >> gr.tf_head(5)
        # straight summarize
        t = df >> gr.tf_summarize(l=gr.last(X.x))
        df_truth = pd.DataFrame({"l": [4.34]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(l=gr.last(X.x))
        df_truth = pd.DataFrame(
            {"cut": ["Good", "Ideal", "Premium"], "l": [4.34, 3.95, 4.20]}
        )
        self.assertTrue(t.equals(df_truth))
        # summarize with order_by
        t = df >> gr.tf_summarize(
            f=gr.last(X.x, order_by=[gr.desc(X.cut), gr.desc(X.x)])
        )
        df_truth = pd.DataFrame({"f": [4.05]})
        assert df_truth.equals(t)
        # straight mutate
        t = df >> gr.tf_mutate(l=gr.last(X.x))
        df_truth = df.copy()
        df_truth["l"] = df_truth.x.iloc[4]
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(l=gr.last(X.x))
        df_truth["l"] = pd.Series([3.95, 4.20, 4.34, 4.20, 4.34])
        self.assertTrue(t.sort_index().equals(df_truth))

    def test_nth(self):
        df = data.df_diamonds >> gr.tf_select(X.cut, X.x) >> gr.tf_head(10)
        # straight summarize
        t = df >> gr.tf_summarize(second=gr.nth(X.x, 1))
        df_truth = pd.DataFrame({"second": [3.89]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(first=gr.nth(X.x, 0))
        df_truth = pd.DataFrame(
            {
                "cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"],
                "first": [3.87, 4.05, 3.95, 3.89, 3.94],
            }
        )
        self.assertTrue(t.equals(df_truth))
        # summarize with order_by
        t = df >> gr.tf_summarize(
            last=gr.nth(X.x, -1, order_by=[gr.desc(X.cut), gr.desc(X.x)])
        )
        df_truth = pd.DataFrame({"last": [3.87]})
        self.assertTrue(df_truth.equals(t))
        # straight mutate
        t = df >> gr.tf_mutate(out_of_range=gr.nth(X.x, 500))
        df_truth = df.copy()
        df_truth["out_of_range"] = np.nan
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(penultimate=gr.nth(X.x, -2))
        df_truth = df.copy()
        df_truth["penultimate"] = pd.Series(
            [np.nan, 3.89, 4.05, 3.89, 4.05, 4.07, 4.07, 4.07, np.nan, 4.07]
        )
        self.assertTrue(t.sort_index().equals(df_truth))

    def test_n(self):
        df = data.df_diamonds >> gr.tf_select(X.cut, X.x) >> gr.tf_head(5)
        # straight summarize
        t = df >> gr.tf_summarize(n=gr.n(X.x))
        df_truth = pd.DataFrame({"n": [5]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(n=gr.n(X.x))
        df_truth = pd.DataFrame({"cut": ["Good", "Ideal", "Premium"], "n": [2, 1, 2]})
        self.assertTrue(t.equals(df_truth))
        # straight mutate
        t = df >> gr.tf_mutate(n=gr.n(X.x))
        df_truth = df.copy()
        df_truth["n"] = 5
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(n=gr.n(X.x))
        df_truth["n"] = pd.Series([1, 2, 2, 2, 2, 2])
        self.assertTrue(t.sort_index().equals(df_truth))
        # Implicit mode summarize
        t = df >> gr.tf_summarize(n=gr.n())
        df_truth = pd.DataFrame({"n": [5]})
        self.assertTrue(t.equals(df_truth))
        # Implicit mode mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(n=gr.n())
        df_truth = df.copy()
        df_truth["n"] = pd.Series([1, 2, 2, 2, 2, 2])
        self.assertTrue(t.sort_index().equals(df_truth))

    def test_n_distinct(self):
        df = pd.DataFrame(
            {
                "col_1": ["a", "a", "a", "b", "b", "b", "c", "c"],
                "col_2": [1, 1, 1, 2, 3, 3, 4, 5],
            }
        )
        # straight summarize
        t = df >> gr.tf_summarize(n=gr.n_distinct(X.col_2))
        df_truth = pd.DataFrame({"n": [5]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.col_1) >> gr.tf_summarize(n=gr.n_distinct(X.col_2))
        df_truth = pd.DataFrame({"col_1": ["a", "b", "c"], "n": [1, 2, 2]})
        self.assertTrue(t.equals(df_truth))
        # straight mutate
        t = df >> gr.tf_mutate(n=gr.n_distinct(X.col_2))
        df_truth = df.copy()
        df_truth["n"] = 5
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.col_1) >> gr.tf_mutate(n=gr.n_distinct(X.col_2))
        df_truth["n"] = pd.Series([1, 1, 1, 2, 2, 2, 2, 2])
        self.assertTrue(t.equals(df_truth))

    def test_IQR(self):
        df = data.df_diamonds >> gr.tf_select(X.cut, X.x) >> gr.tf_head(5)
        # straight summarize
        t = df >> gr.tf_summarize(i=gr.IQR(X.x))
        df_truth = pd.DataFrame({"i": [0.25]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(i=gr.IQR(X.x))
        df_truth = pd.DataFrame(
            {"cut": ["Good", "Ideal", "Premium"], "i": [0.145, 0.000, 0.155]}
        )
        test_vector = abs(t.i - df_truth.i)
        assert all(test_vector < 0.000000001)
        # straight mutate
        t = df >> gr.tf_mutate(i=gr.IQR(X.x))
        df_truth = df.copy()
        df_truth["i"] = 0.25
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(i=gr.IQR(X.x))
        df_truth["i"] = pd.Series([0.000, 0.155, 0.145, 0.155, 0.145])
        test_vector = abs(t.i - df_truth.i)
        self.assertTrue(all(test_vector < 0.000000001))

    def test_min(self):
        df = data.df_diamonds >> gr.tf_select(X.cut, X.x) >> gr.tf_head(5)
        # straight summarize
        t = df >> gr.tf_summarize(m=gr.min(X.x))
        df_truth = pd.DataFrame({"m": [3.89]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(m=gr.min(X.x))
        df_truth = pd.DataFrame(
            {"cut": ["Good", "Ideal", "Premium"], "m": [4.05, 3.95, 3.89]}
        )
        self.assertTrue(t.equals(df_truth))
        # straight mutate
        t = df >> gr.tf_mutate(m=gr.min(X.x))
        df_truth = df.copy()
        df_truth["m"] = 3.89
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(m=gr.min(X.x))
        df_truth["m"] = pd.Series([3.95, 3.89, 4.05, 3.89, 4.05])
        self.assertTrue(t.sort_index().equals(df_truth))

    def test_max(self):
        df = data.df_diamonds >> gr.tf_select(X.cut, X.x) >> gr.tf_head(5)
        # straight summarize
        t = df >> gr.tf_summarize(m=gr.max(X.x))
        df_truth = pd.DataFrame({"m": [4.34]})
        self.assertTrue(t.equals(df_truth))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(m=gr.max(X.x))
        df_truth = pd.DataFrame(
            {"cut": ["Good", "Ideal", "Premium"], "m": [4.34, 3.95, 4.20]}
        )
        self.assertTrue(t.equals(df_truth))
        # straight mutate
        t = df >> gr.tf_mutate(m=gr.max(X.x))
        df_truth = df.copy()
        df_truth["m"] = 4.34
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(m=gr.max(X.x))
        df_truth["m"] = pd.Series([3.95, 4.20, 4.34, 4.20, 4.34])
        self.assertTrue(t.sort_index().equals(df_truth))

    def test_median(self):
        df = (
            data.df_diamonds
            >> gr.tf_group_by(X.cut)
            >> gr.tf_head(3)
            >> gr.tf_select(X.cut, X.x)
            >> gr.tf_ungroup()
        )
        # straight summarize
        t = df >> gr.tf_summarize(m=gr.median(X.x))
        df_truth = pd.DataFrame({"m": [4.05]})
        self.assertTrue(t.equals(df_truth))

        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(m=gr.median(X.x))
        df_truth = pd.DataFrame(
            {
                "cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"],
                "m": [6.27, 4.25, 3.95, 3.89, 3.95],
            }
        )
        self.assertTrue(t.equals(df_truth))
        # straight mutate
        t = df >> gr.tf_mutate(m=gr.median(X.x))
        df_truth = df.copy()
        df_truth["m"] = 4.05
        self.assertTrue(t.equals(df_truth))
        # grouped mutate
        # t = df >> group_by(X.cut) >> mutate(m=median(X.x))
        # df_truth['m'] = pd.Series(
        #     [6.27, 6.27, 6.27, 4.25, 4.25, 4.25, 3.95, 3.95, 3.95, 3.89, 3.89, 3.89, 3.95, 3.95, 3.95],
        #     index=t.index)
        # assert t.equals(df_truth)
        # make sure it handles case with even counts properly
        df = (
            data.df_diamonds
            >> gr.tf_group_by(X.cut)
            >> gr.tf_head(2)
            >> gr.tf_select(X.cut, X.x)
        )
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(m=gr.median(X.x))
        df_truth = pd.DataFrame(
            {
                "cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"],
                "m": [5.160, 4.195, 3.940, 4.045, 3.945],
            }
        )
        test_vector = abs(t.m - df_truth.m)
        self.assertTrue(all(test_vector < 0.000000001))

    def test_var(self):
        df = (
            data.df_diamonds
            >> gr.tf_group_by(X.cut)
            >> gr.tf_head(3)
            >> gr.tf_select(X.cut, X.x)
            >> gr.tf_ungroup()
        )

        # straight summarize
        t = df >> gr.tf_summarize(v=gr.var(X.x))
        df_truth = pd.DataFrame({"v": [0.687392]})
        test_vector = abs(t.v - df_truth.v)
        self.assertTrue(all(test_vector < 0.00001))

        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(v=gr.var(X.x))
        df_truth = pd.DataFrame(
            {
                "cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"],
                "v": [2.074800, 0.022033, 0.056133, 0.033100, 0.005233],
            }
        )
        test_vector = abs(t.v - df_truth.v)
        self.assertTrue(all(test_vector < 0.00001))
        # straight mutate
        t = df >> gr.tf_mutate(v=gr.var(X.x))
        df_truth = df.copy()
        df_truth["v"] = 0.687392
        test_vector = abs(t.v - df_truth.v)
        self.assertTrue(all(test_vector < 0.00001))
        # grouped mutate
        # t = df >> group_by(X.cut) >> mutate(v=var(X.x))
        # df_truth['v'] = pd.Series([2.074800, 2.074800, 2.074800, 0.022033, 0.022033, 0.022033,
        #                            0.056133, 0.056133, 0.056133, 0.033100, 0.033100, 0.033100,
        #                            0.005233, 0.005233, 0.005233],
        #                           index=t.index)
        # test_vector = abs(t.v - df_truth.v)
        # assert all(test_vector < .00001)
        # test with single value (var undefined)
        df = (
            data.df_diamonds
            >> gr.tf_group_by(X.cut)
            >> gr.tf_head(1)
            >> gr.tf_select(X.cut, X.x)
        )
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(v=gr.var(X.x))
        df_truth = pd.DataFrame(
            {
                "cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"],
                "v": [np.nan, np.nan, np.nan, np.nan, np.nan],
            }
        )
        self.assertTrue(t.equals(df_truth))

    def test_sd(self):
        df = (
            data.df_diamonds
            >> gr.tf_group_by(X.cut)
            >> gr.tf_head(3)
            >> gr.tf_select(X.cut, X.x)
            >> gr.tf_ungroup()
        )
        # straight summarize
        t = df >> gr.tf_summarize(s=gr.sd(X.x))
        df_truth = pd.DataFrame({"s": [0.829091]})
        test_vector = abs(t.s - df_truth.s)
        self.assertTrue(all(test_vector < 0.00001))
        # grouped summarize
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(s=gr.sd(X.x))
        df_truth = pd.DataFrame(
            {
                "cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"],
                "s": [1.440417, 0.148436, 0.236925, 0.181934, 0.072342],
            }
        )
        test_vector = abs(t.s - df_truth.s)
        self.assertTrue(all(test_vector < 0.00001))
        # straight mutate
        t = df >> gr.tf_mutate(s=gr.sd(X.x))
        df_truth = df.copy()
        df_truth["s"] = 0.829091
        test_vector = abs(t.s - df_truth.s)
        self.assertTrue(all(test_vector < 0.00001))
        # grouped mutate
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_mutate(s=gr.sd(X.x))
        # df_truth['s'] = pd.Series([1.440417, 1.440417, 1.440417, 0.148436, 0.148436, 0.148436,
        #                            0.236925, 0.236925, 0.236925, 0.181934, 0.181934, 0.181934,
        #                            0.072342, 0.072342, 0.072342],
        #                           index=t.index)
        # test_vector = abs(t.s - df_truth.s)
        # print(t)
        # print(df_truth)
        self.assertTrue(all(test_vector < 0.00001))
        # test with single value (var undefined)
        df = (
            data.df_diamonds
            >> gr.tf_group_by(X.cut)
            >> gr.tf_head(1)
            >> gr.tf_select(X.cut, X.x)
        )
        t = df >> gr.tf_group_by(X.cut) >> gr.tf_summarize(s=gr.sd(X.x))
        df_truth = pd.DataFrame(
            {
                "cut": ["Fair", "Good", "Ideal", "Premium", "Very Good"],
                "s": [np.nan, np.nan, np.nan, np.nan, np.nan],
            }
        )
        self.assertTrue(t.equals(df_truth))

    def test_quant(self):
        df = pd.DataFrame(data={"x": [0, 0.25, 0.5, 0.75, 1]})

        df_t_25 = pd.DataFrame({"q": [0.25]})
        df_t_50 = pd.DataFrame({"q": [0.50]})
        df_t_75 = pd.DataFrame({"q": [0.75]})

        df_c_25 = df >> gr.tf_summarize(q=gr.quant(X.x, p=0.25))
        df_c_50 = df >> gr.tf_summarize(q=gr.quant(X.x, p=0.50))
        df_c_75 = df >> gr.tf_summarize(q=gr.quant(X.x, p=0.75))

        self.assertTrue(df_t_25.equals(df_c_25))
        self.assertTrue(df_t_50.equals(df_c_50))
        self.assertTrue(df_t_75.equals(df_c_75))

    def test_rsq(self):
        # Known case; Rsq = 3/4
        y_meas = pd.Series([-1, 0, +1])
        y_fit = pd.Series([-0.5, 0, +0.5])
        rsq_comp = gr.rsq(y_fit, y_meas)

        self.assertTrue(rsq_comp, 3 / 4)

    def test_corr(self):
        df_data = gr.df_make(x=[1., 2., 3., 4.])
        df_data["y"] = 0.5 * df_data.x
        df_data["z"] = - 0.5 * df_data.x

        self.assertTrue(abs(gr.corr(df_data.x, df_data.y) - 1.0) < 1e-6)
        self.assertTrue(abs(gr.corr(df_data.x, df_data.z) + 1.0) < 1e-6)

        ## Test NaN handling
        df_nan = (
            df_data
            >> gr.tf_mutate(
                x=gr.if_else(X.x == 1, gr.NaN, X.x),
                y=gr.if_else(X.x == 4, gr.NaN, X.y),
            )
        )

        with self.assertRaises(ValueError):
            gr.corr(df_nan.x, df_nan.y)
        self.assertTrue(abs(gr.corr(df_nan.x, df_nan.y, nan_drop=True) - 1.0) < 1e-6)

class TestCIHelpers(unittest.TestCase):

    def test_mean_ci(self):
        # Basic functionality
        y = pd.Series([-1, -1, 0, +1, +1]) # sd == 1
        lo_true = 0 - (-norm.ppf(0.005)) * 1 / np.sqrt(5)
        up_true = 0 + (-norm.ppf(0.005)) * 1 / np.sqrt(5)

        self.assertTrue((lo_true - gr.mean_lo(y, alpha=0.005)) < 1e-6)
        self.assertTrue((up_true - gr.mean_up(y, alpha=0.005)) < 1e-6)

        # Grouped functionality
        df = (
            gr.df_grid(
                y=[-1, -1, 0, +1, +1],
                x=[0, 1],
            )
            >> gr.tf_mutate(y=X.y + X.x)
            >> gr.tf_group_by(X.x)
            >> gr.tf_summarize(
                mean_lo=gr.mean_lo(X.y),
                mean_up=gr.mean_up(X.y),
            )
        )

        self.assertTrue(
            (df[df.x==0].mean_lo.values[0] - lo_true) < 1e-6
        )
        self.assertTrue(
            (df[df.x==0].mean_up.values[0] - up_true) < 1e-6
        )

        self.assertTrue(
            (df[df.x==1].mean_lo.values[0] - (lo_true + 1)) < 1e-6
        )
        self.assertTrue(
            (df[df.x==1].mean_up.values[0] - (up_true + 1)) < 1e-6
        )


    def test_pr_ci(self):
        # Basic functionality
        t = pd.Series([1, 1, 1])
        f = pd.Series([0, 0, 0])

        # Wilson intervals respect 0, 1 bounds
        self.assertTrue(gr.pr_lo(f) > -1e-6)
        self.assertTrue(gr.pr_up(t) < 1 + 1e-6)
        # Correct ordering
        self.assertTrue(gr.pr_lo(t) <= gr.pr(t))
        self.assertTrue(gr.pr(t) <= gr.pr_up(t))
        self.assertTrue(gr.pr_lo(f) <= gr.pr(f))
        self.assertTrue(gr.pr(f) <= gr.pr_up(f))


    def test_prediction_intervals(self):
        ## Correct indexes
        # Example 5.11, Hahn and Meeker
        idx = gr.pint_up_index(100, 59, 30, 0.05)
        self.assertTrue(idx == 64)
        # Example 5.12, Hahn and Meeker
        idx = gr.pint_lo_index(100, 59, 30, 0.05)
        self.assertTrue(idx == 37)

        ## Test functionality
        df_res = (
            data.df_shewhart
            >> gr.tf_summarize(
                pi_lo=gr.pint_lo(X.tensile_strength, alpha=0.10/2),
                pi_up=gr.pint_up(X.tensile_strength, alpha=0.10/2),
            )
        )
        # Raises assertion
        with self.assertRaises(ValueError):
            df_res = (
                data.df_shewhart
                >> gr.tf_summarize(
                    pi_lo=gr.pint_lo(X.tensile_strength),
                    pi_up=gr.pint_up(X.tensile_strength),
                )
            )



import numpy as np
import pandas as pd
import unittest

from collections import OrderedDict as od
from context import core
from context import grama as gr
from context import models
from pyDOE import lhs

DF = gr.Intention()

##################################################
class TestEvalInvariants(unittest.TestCase):
    """Helper class for testing invariant errors for eval_* functions"""
    def __init__(self):
        """Setup necessary values"""
        self.md = (
            gr.Model()
            >> gr.cp_function(fun=lambda x: x, var=1, out=1, runtime=1)
        )
        self.md_var_det = self.md >> gr.cp_bounds(x1=(0, 1))
        self.df = pd.DataFrame(
            data={"x": [0.0], "y": [0.5]}
        )
        # declare tests
        self.type_tests = [(1,2), 2, [1, 8]]

    def md_arg(self, func, df_arg = "df", **kwargs):
        """Helper function for TypeErrors and ValueErrors for invalid
        Model arguments (eval_* functions).

        Args:
            func (func): eval function to test
            df_arg (str): name of DataFrame argument
            **kwargs: kwargs to pass"""

        ## Type test
        for wrong in self.type_tests:
            self.assertRaises(TypeError, func, wrong, **{df_arg:self.df}, **kwargs)

        ## No model.functions
        self.assertRaises(ValueError, func, gr.Model(), **{df_arg:self.df}, **kwargs)

    def df_arg(self, func, df_arg = "df", shortcut=False, acc_none="Never", **kwargs):
        """Helper function for testing for TypeErrors and ValueErrors for invalid
        DataFrame arguments (eval_* functions).

        Args:
            func (func): eval function to test
            df_arg (str): name of DataFrame argument
            shortcut (bool): if func has valid str shortcut for df arg
            acc_none (str or None): if func accepts None for df;
                "var_det": accepts None when model.n_var_det == 0;
                "always": always accepts None as input
        """
        ## General type tests
        for wrong in self.type_tests:
            self.assertRaises(TypeError, func, self.md, **{df_arg:wrong}, **kwargs)

        ## Str type test
        if shortcut:
            # wrong str shortcut test
            self.assertRaises(ValueError, func, self.md, **{df_arg:"a"}, **kwargs)
        else:
            # test any str for typeerror
            self.assertRaises(TypeError, func, self.md, **{df_arg:"nom"}, **kwargs)

        ## None checks
        if acc_none == "var_det" or acc_none == "never":
            # `None` check when model.n_var_det > 0
            self.assertRaises(TypeError, func, self.md_var_det, **{df_arg:None}, **kwargs)
        if acc_none == "never":
            # none not accepted under any condition, test when md.var_det==0
            self.assertRaises(TypeError, func, self.md, **{df_arg:None}, **kwargs)

    def df_arg_2(self, func, df_args,
                    shortcut=[False, False], acc_none=["Never", "Never"],
                    **kwargs):
        """Helper function for testing for TypeErrors and ValueErrors for
        DataFrame arguments in eval_* functions with two df inputs.

        Args:
            func (func): eval function to test
            df_arg1 (list(str)): name of DataFrame args
            shortcut (list(bool)): if func has valid str shortcut for either or
                both df args
            acc_none (list(str) or list(None)): if func accepts None for df args;
                "var_det": accepts None when model.n_var_det == 0;
                "always": always accepts None as input
                "never": never accepts None as input
            """
        ## General type tests for both dfs
        for wrong in self.type_tests:
            self.assertRaises(TypeError, func, self.md,
                                **{df_args[0]:wrong}, **{df_args[1]:self.df},
                                **kwargs)
            self.assertRaises(TypeError, func, self.md,
                                **{df_args[0]:self.df}, **{df_args[1]:wrong,
                                **kwargs})

        ## Str type test
        # df_args[0]
        if shortcut[0]:
            # wrong str shortcut test
            self.assertRaises(ValueError, func, self.md,
                                **{df_args[0]:"a"}, **{df_args[1]:self.df},
                                **kwargs)
        else:
            # test any str for typeerror
            self.assertRaises(TypeError, func, self.md,
                                **{df_args[0]:"nom"}, **{df_args[1]:self.df},
                                **kwargs)
        # df_args[1]
        if shortcut[1]:
            # wrong str shortcut test
            self.assertRaises(ValueError, func, self.md,
                                **{df_args[0]:self.df}, **{df_args[1]:"a"},
                                **kwargs)
        else:
            # test any str for typeerror
            self.assertRaises(TypeError, func, self.md,
                                **{df_args[0]:self.df}, **{df_args[1]:"nom"},
                                **kwargs)

        ## None checks
        if acc_none[0] == "var_det" or acc_none[0] is None:
            # `None` check when model.n_var_det > 0
            self.assertRaises(TypeError, func, self.md_var_det,
                                **{df_args[0]:None}, **{df_args[1]:self.df},
                                **kwargs)
        if acc_none[0] is None:
            # none not accepted under any condition, test when md.var_det==0
            self.assertRaises(TypeError, func, self.md,
                                **{df_args[0]:None}, **{df_args[1]:self.df},
                                **kwargs)
        # df_arg_2
        if acc_none[1] == "var_det" or acc_none[1] is None:
            # `None` check when model.n_var_det > 0
            self.assertRaises(TypeError, func, self.md_var_det,
                                **{df_args[0]:self.df}, **{df_args[1]:None},
                                **kwargs)
        if acc_none[1] is None:
            # none not accepted under any condition, test when md.var_det==0
            self.assertRaises(TypeError, func, self.md,
                                **{df_args[0]:self.df}, **{df_args[1]:None},
                                **kwargs)

##################################################
class TestDefaults(unittest.TestCase):
    def setUp(self):
        # 2D identity model with permuted df inputs
        domain_2d = gr.Domain(bounds={"x": [-1.0, +1], "y": [0.0, 1.0]})
        marginals = {}
        marginals["x"] = gr.MarginalNamed(
            d_name="uniform", d_param={"loc": -1, "scale": 2}
        )
        marginals["y"] = gr.MarginalNamed(
            sign=-1, d_name="uniform", d_param={"loc": 0, "scale": 1}
        )

        self.model_2d = gr.Model(
            functions=[
                gr.Function(lambda x, y: [x, y], ["x", "y"], ["f", "g"], "test", 0)
            ],
            domain=domain_2d,
            density=gr.Density(
                marginals=marginals, copula=gr.CopulaIndependence(var_rand=["x"])
            ),
        )

        ## Invariant test class
        self.inv_test = TestEvalInvariants()

        ## Correct results
        self.df_2d_nominal = pd.DataFrame(
            data={"x": [0.0], "y": [0.5], "f": [0.0], "g": [0.5]}
        )
        self.df_2d_grad = pd.DataFrame(
            data={"Df_Dx": [1.0], "Dg_Dx": [0.0], "Df_Dy": [0.0], "Dg_Dy": [1.0]}
        )
        self.df_2d_qe = pd.DataFrame(
            data={"x": [0.0], "y": [0.1], "f": [0.0], "g": [0.1]}
        )

    def test_nominal(self):
        """Checks the nominal evaluation is accurate
        """
        df_res = gr.eval_nominal(self.model_2d)

        ## Accurate
        self.assertTrue(gr.df_equal(self.df_2d_nominal, df_res))

        ## Invariant checks
        self.inv_test.md_arg(gr.eval_nominal, df_arg="df_det")
        self.inv_test.df_arg(gr.eval_nominal, df_arg="df_det", shortcut=True, acc_none="var_det")

        ## Pass-through
        self.assertTrue(
            gr.df_equal(
                self.df_2d_nominal.drop(["f", "g"], axis=1),
                gr.eval_nominal(self.model_2d, skip=True),
            )
        )

    def test_linup(self):
        """Checks linear uncertainty propagation tool
        """
        # Setup
        model_3d = (
            gr.Model()
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(
                    y0=df.x0 + df.x1 + df.x2,
                    y1=df.x0 * gr.exp(df.x1 + df.x2),
                ),
                var=3,
                out=2
            )
            >> gr.cp_bounds(x0=(-1, +1))
            >> gr.cp_marginals(
                x1=gr.marg_mom("norm", mean=0, sd=1),
                x2=gr.marg_mom("uniform", mean=0, sd=1),
            )
            >> gr.cp_copula_gaussian(df_corr=gr.df_make(var1=["x1"], var2=["x2"], corr=[0.5]))
        )

        model_50 = (
            gr.Model()
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(
                    y0=df.x0 + df.x1 + df.x2,
                ),
                var=3,
                out=1,
            )
            >> gr.cp_bounds(x0=(-1, +1))
            >> gr.cp_marginals(
                x1=gr.marg_mom("norm", mean=0, sd=1),
                x2=gr.marg_mom("norm", mean=0, sd=1),
            )
            >> gr.cp_copula_gaussian(df_corr=gr.df_make(var1=["x1"], var2=["x2"], corr=[0.5]))
        )
        model_75 = (
            gr.Model()
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(
                    y0=df.x0 + df.x1 + df.x2,
                ),
                var=3,
                out=1,
            )
            >> gr.cp_bounds(x0=(-1, +1))
            >> gr.cp_marginals(
                x1=gr.marg_mom("norm", mean=0, sd=1),
                x2=gr.marg_mom("norm", mean=0, sd=1),
            )
            >> gr.cp_copula_gaussian(df_corr=gr.df_make(var1=["x1"], var2=["x2"], corr=[0.75]))
        )

        ## Test for accuracy
        df_res_50 = gr.eval_linup(model_50, df_base="nom", n=1e5, seed=101)
        self.assertTrue(abs(df_res_50["var"].values[0] - 3) / 3 <= 0.05)
        df_res_75 = gr.eval_linup(model_75, df_base="nom", n=1e5, seed=101)
        self.assertTrue(abs(df_res_75["var"].values[0] - 3.5) / 3.5 <= 0.05)

        ## Test for evaluation at base values
        df_base = gr.df_make(x0=0, x1=-1, x2=1)
        df_3d_base = gr.eval_linup(model_3d, df_base=df_base)
        self.assertTrue(gr.df_equal(
            pd.concat((df_base, df_base), axis=0).reset_index(drop=True),
            df_3d_base[df_base.columns],
        ))

        ## Test for outer product of base and outputs
        df_3d_outer = gr.eval_linup(model_3d, df_base=gr.df_make(x0=0, x1=0, x2=[0, 1]))
        self.assertTrue(df_3d_outer.shape[0] == 2 * 2)

        ## Test that sensitivity contributions sum to 1
        df_3d_nom = gr.eval_linup(model_3d, df_base=gr.df_make(x0=1, x1=0, x2=0), decomp=True)
        df_sum = (
            df_3d_nom
            >> gr.tf_group_by("out")
            >> gr.tf_summarize(s=gr.sum(DF.var_frac))
        )
        self.assertTrue(all(np.abs(df_sum.s - 1) <= 0.01))


    def test_grad_fd(self):
        """Checks the FD code
        """
        ## Accuracy
        df_grad = gr.eval_grad_fd(
            self.model_2d, df_base=self.df_2d_nominal, append=False
        )

        self.assertTrue(np.allclose(df_grad[self.df_2d_grad.columns], self.df_2d_grad))

        ## Invariant checks
        self.inv_test.md_arg(gr.eval_grad_fd, df_arg="df_base")
        self.inv_test.df_arg(gr.eval_grad_fd, df_arg="df_base")

        ## Subset
        df_grad_sub = gr.eval_grad_fd(
            self.model_2d, df_base=self.df_2d_nominal, var=["x"], append=False
        )

        self.assertTrue(set(df_grad_sub.columns) == set(["Df_Dx", "Dg_Dx"]))

        ## Flags
        md_test = (
            gr.Model()
            >> gr.cp_function(fun=lambda x0, x1: x0 + x1 ** 2, var=2, out=1)
            >> gr.cp_marginals(x0={"dist": "norm", "loc": 0, "scale": 1})
        )
        df_base = pd.DataFrame(dict(x0=[0, 1], x1=[0, 1]))
        ## Multiple base points
        df_true = pd.DataFrame(dict(Dy0_Dx0=[1, 1], Dy0_Dx1=[0, 2]))

        df_rand = gr.eval_grad_fd(md_test, df_base=df_base, var="rand", append=False)
        self.assertTrue(gr.df_equal(df_true[["Dy0_Dx0"]], df_rand, close=True))

        df_det = gr.eval_grad_fd(md_test, df_base=df_base, var="det", append=False)
        self.assertTrue(gr.df_equal(df_true[["Dy0_Dx1"]], df_det, close=True))
        ## Append base points
        df_append = gr.eval_grad_fd(md_test, df_base=df_base, var="det", append=True)
        self.assertTrue(gr.df_equal(df_base, df_append[md_test.var]))

    def test_conservative(self):
        ## Accuracy
        df_res = gr.eval_conservative(self.model_2d, quantiles=[0.1, 0.1])

        self.assertTrue(gr.df_equal(self.df_2d_qe, df_res, close=True))

        ## Invariant checks
        self.inv_test.md_arg(gr.eval_conservative, df_arg="df_det")
        self.inv_test.df_arg(gr.eval_conservative, df_arg="df_det",
                                shortcut=True, acc_none="var_det")

        ## Repeat scalar value
        self.assertTrue(
            gr.df_equal(
                self.df_2d_qe,
                gr.eval_conservative(self.model_2d, quantiles=0.1),
                close=True,
            )
        )

        ## Pass-through
        self.assertTrue(
            gr.df_equal(
                self.df_2d_qe.drop(["f", "g"], axis=1),
                gr.eval_conservative(self.model_2d, quantiles=0.1, skip=True),
                close=True,
            )
        )


##################################################
class TestRandomSampling(unittest.TestCase):
    def setUp(self):
        self.md = (
            gr.Model()
            >> gr.cp_function(fun=lambda x: x, var=1, out=1, runtime=1)
            >> gr.cp_marginals(x0={"dist": "uniform", "loc": 0, "scale": 1})
            >> gr.cp_copula_independence()
        )

        self.md_2d = (
            gr.Model()
            >> gr.cp_function(fun=lambda x0, x1: x0, var=2, out=1)
            >> gr.cp_marginals(
                x0={"dist": "uniform", "loc": 0, "scale": 1},
                x1={"dist": "uniform", "loc": 0, "scale": 1},
            )
            >> gr.cp_copula_independence()
        )

        self.md_mixed = (
            gr.Model()
            >> gr.cp_function(fun=lambda x0, x1: x0 + x1, var=2, out=1)
            >> gr.cp_bounds(x0=(-1, +1))
            >> gr.cp_marginals(
                x1={"dist": "uniform", "loc": 0, "scale": 1},
            )
            >> gr.cp_copula_independence()
        )

    def test_lhs(self):
        ## Accurate
        n = 2
        df_res = gr.eval_lhs(self.md_2d, n=n, df_det="nom", seed=101)

        np.random.seed(101)
        df_truth = pd.DataFrame(data=lhs(2, samples=n), columns=["x0", "x1"])
        df_truth["y0"] = df_truth["x0"]

        self.assertTrue(gr.df_equal(df_res, df_truth))

        ## Rounding
        df_round = gr.eval_lhs(self.md_2d, n=n + 0.1, df_det="nom", seed=101)

        self.assertTrue(gr.df_equal(df_round, df_truth))

        ## Pass-through
        df_pass = gr.eval_lhs(self.md_2d, n=n, skip=True, df_det="nom", seed=101)

        self.assertTrue(gr.df_equal(df_pass, df_truth[["x0", "x1"]]))

    def test_sample(self):
        ## Accurate
        n = 2
        df_res = gr.eval_sample(self.md, n=n, df_det="nom", seed=101)

        np.random.seed(101)
        df_truth = pd.DataFrame({"x0": np.random.random(n)})
        df_truth["y0"] = df_truth["x0"]

        self.assertTrue(gr.df_equal(df_res, df_truth))

        ## Rounding
        df_round = gr.eval_sample(self.md, n=n + 0.1, df_det="nom", seed=101)

        self.assertTrue(gr.df_equal(df_round, df_truth))

        ## Pass-through
        df_pass = gr.eval_sample(self.md, n=n, skip=True, df_det="nom", seed=101)

        self.assertTrue(gr.df_equal(df_pass[["x0"]], df_truth[["x0"]]))

        ## Optional observation index
        df_idx = gr.eval_sample(
            self.md_mixed,
            n=n,
            df_det=gr.df_make(x0=[-1, 0, 1]),
            seed=101,
            ind_comm="idx",
        )

        self.assertTrue(len(set(df_idx.idx)) == n)

        ## Common random numbers
        df_crn = gr.eval_sample(
            self.md_mixed,
            n=3,
            df_det=gr.df_make(x0=[0, 1]),
            seed=101,
        )
        self.assertTrue(len(set(df_crn.x1)) == 3)

        ## Non-common random numbers
        df_ncrn = gr.eval_sample(
            self.md_mixed,
            n=3,
            df_det=gr.df_make(x0=[0, 1]),
            seed=101,
            comm=False,
        )
        self.assertTrue(len(set(df_ncrn.x1)) == 6)


##################################################
class TestRandom(unittest.TestCase):
    def setUp(self):
        self.md = models.make_test()

        self.md_mixed = (
            gr.Model()
            >> gr.cp_function(fun=lambda x0, x1, x2: x0, var=3, out=1)
            >> gr.cp_bounds(x2=(0, 1))
            >> gr.cp_marginals(
                x0={"dist": "uniform", "loc": 0, "scale": 1},
                x1={"dist": "uniform", "loc": 0, "scale": 1},
            )
            >> gr.cp_copula_independence()
        )

        ## Invariant test class
        self.inv_test = TestEvalInvariants()

    def test_sample(self):
        # invariant checks
        self.inv_test.md_arg(gr.eval_sample, df_arg="df_det")
        self.inv_test.df_arg(gr.eval_sample, df_arg="df_det", shortcut=True,
                                    acc_none="var_det")

        # No `n` provided
        with self.assertRaises(ValueError):
            gr.eval_sample(self.md, df_det="nom")

        df_min = gr.eval_sample(self.md, n=1, df_det="nom")
        self.assertTrue(df_min.shape == (1, self.md.n_var + self.md.n_out))
        self.assertTrue(set(df_min.columns) == set(self.md.var + self.md.out))

        # Seed fixes runs
        df_seeded = gr.eval_sample(self.md, n=10, df_det="nom", seed=101)
        df_piped = self.md >> gr.ev_sample(df_det="nom", n=10, seed=101)
        self.assertTrue(df_seeded.equals(df_piped))

        df_skip = gr.eval_sample(self.md, n=1, df_det="nom", skip=True)
        self.assertTrue(set(df_skip.columns) == set(self.md.var))

        df_noappend = gr.eval_sample(self.md, n=1, df_det="nom", append=False)
        self.assertTrue(set(df_noappend.columns) == set(self.md.out))

    def test_lhs(self):
        # no inv test implemented, no need to test

        df_seeded = gr.eval_lhs(self.md, n=10, df_det="nom", seed=101)
        df_piped = self.md >> gr.ev_lhs(df_det="nom", n=10, seed=101)
        self.assertTrue(df_seeded.equals(df_piped))

        df_skip = gr.eval_lhs(self.md, n=1, df_det="nom", skip=True)
        self.assertTrue(set(df_skip.columns) == set(self.md.var))

        df_noappend = gr.eval_lhs(self.md, n=1, df_det="nom", append=False)
        self.assertTrue(set(df_noappend.columns) == set(self.md.out))

    def test_sinews(self):
        # invariant checks
        self.inv_test.md_arg(gr.eval_sinews, df_arg="df_det")
        self.inv_test.df_arg(gr.eval_sinews, df_arg="df_det", shortcut=True)

        df_min = gr.eval_sinews(self.md, df_det="nom")
        self.assertTrue(
            set(df_min.columns)
            == set(self.md.var + self.md.out + ["sweep_var", "sweep_ind"])
        )
        self.assertTrue(df_min._plot_info["type"] == "sinew_outputs")

        df_seeded = gr.eval_sinews(self.md, df_det="nom", seed=101)
        df_piped = self.md >> gr.ev_sinews(df_det="nom", seed=101)
        self.assertTrue(df_seeded.equals(df_piped))

        df_skip = gr.eval_sinews(self.md, df_det="nom", skip=True)
        self.assertTrue(df_skip._plot_info["type"] == "sinew_inputs")

        df_mixed = gr.eval_sinews(self.md_mixed, df_det="swp")

    def test_hybrid(self):
        # invariant checks
        self.inv_test.md_arg(gr.eval_hybrid, df_arg="df_det")
        self.inv_test.df_arg(gr.eval_hybrid, df_arg="df_det", shortcut=True)

        df_min = gr.eval_hybrid(self.md, df_det="nom")
        self.assertTrue(
            set(df_min.columns) == set(self.md.var + self.md.out + ["hybrid_var"])
        )
        self.assertTrue(df_min._meta["type"] == "eval_hybrid")

        df_seeded = gr.eval_hybrid(self.md, df_det="nom", seed=101)
        df_piped = self.md >> gr.ev_hybrid(df_det="nom", seed=101)
        self.assertTrue(df_seeded.equals(df_piped))

        df_total = gr.eval_hybrid(self.md, df_det="nom", plan="total")
        self.assertTrue(
            set(df_total.columns) == set(self.md.var + self.md.out + ["hybrid_var"])
        )
        self.assertTrue(df_total._meta["type"] == "eval_hybrid")

        df_skip = gr.eval_hybrid(self.md, df_det="nom", skip=True)
        self.assertTrue(set(df_skip.columns) == set(self.md.var + ["hybrid_var"]))

        ## Raises
        md_buckle = models.make_plate_buckle()
        with self.assertRaises(ValueError):
            gr.eval_hybrid(md_buckle, df_det="nom")


##################################################
class TestOpt(unittest.TestCase):
    def setUp(self):
        ## Invariant test class
        self.inv_test = TestEvalInvariants()

    def test_nls(self):
        # invariant checks
        self.inv_test.md_arg(gr.eval_nls, df_arg="df_data")
        self.inv_test.df_arg_2(gr.eval_nls, df_args=["df_data", "df_init"],
            acc_none=["never", "always"])

        ## Setup
        md_feat = (
            gr.Model()
            >> gr.cp_function(fun=lambda x0, x1, x2: x0 * x1 + x2, var=3, out=1,)
            >> gr.cp_bounds(x0=[-1, +1], x2=[0, 0])
            >> gr.cp_marginals(x1=dict(dist="norm", loc=0, scale=1))
        )

        md_const = (
            gr.Model()
            >> gr.cp_function(fun=lambda x0: x0, var=1, out=1)
            >> gr.cp_bounds(x0=(-1, +1))
        )

        df_response = md_feat >> gr.ev_df(
            df=gr.df_make(x0=0.1, x1=[-1, -0.5, +0, +0.5, +1], x2=0)
        )
        df_data = df_response[["x1", "y0"]]

        ## Model with features
        df_true = gr.df_make(x0=0.1)
        df_fit = md_feat >> gr.ev_nls(df_data=df_data, append=False)

        pd.testing.assert_frame_equal(
            df_fit,
            df_true,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )
        ## Fitting synonym
        md_feat_fit = df_data >> gr.ft_nls(md=md_feat, verbose=False)
        self.assertTrue(set(md_feat_fit.var) == set(["x1", "x2"]))

        ## Constant model
        df_const = gr.df_make(x0=0)
        df_fit = md_const >> gr.ev_nls(df_data=gr.df_make(y0=[-1, 0, +1]))

        pd.testing.assert_frame_equal(
            df_fit,
            df_const,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )

        ## Multiple restarts works
        df_multi = gr.eval_nls(md_feat, df_data=df_data, n_restart=2)
        self.assertTrue(df_multi.shape[0] == 2)

        ## Specified initial guess
        df_spec = gr.eval_nls(
            md_feat, df_data=df_data, df_init=gr.df_make(x0=0.5), append=False
        )
        pd.testing.assert_frame_equal(
            df_spec,
            df_true,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )
        # Raises if incorrect guess data
        with self.assertRaises(ValueError):
            gr.eval_nls(md_feat, df_data=df_data, df_init=gr.df_make(foo=0.5))

    def test_opt(self):
        # invariant checks
        self.inv_test.md_arg(gr.eval_min, df_arg="df_start")
        self.inv_test.df_arg(gr.eval_min, df_arg="df_start", acc_none="always")


        md_bowl = (
            gr.Model("Constrained bowl")
            >> gr.cp_function(
                fun=lambda x, y: x ** 2 + y ** 2, var=["x", "y"], out=["f"],
            )
            >> gr.cp_function(
                fun=lambda x, y: (x + y + 1), var=["x", "y"], out=["g1"],
            )
            >> gr.cp_function(
                fun=lambda x, y: -(-x + y - np.sqrt(2 / 10)),
                var=["x", "y"],
                out=["g2"],
            )
            >> gr.cp_bounds(x=(-1, +1), y=(-1, +1),)
        )

        df_res = md_bowl >> gr.ev_min(out_min="f", out_geq=["g1"], out_leq=["g2"],)

        # Check result
        self.assertTrue(abs(df_res.x[0] + np.sqrt(1 / 20)) < 1e-6)
        self.assertTrue(abs(df_res.y[0] - np.sqrt(1 / 20)) < 1e-6)

        # Check errors for violated invariants
        with self.assertRaises(ValueError):
            gr.eval_min(md_bowl, out_min="FALSE")
        with self.assertRaises(ValueError):
            gr.eval_min(md_bowl, out_min="f", out_geq=["FALSE"])
        with self.assertRaises(ValueError):
            gr.eval_min(md_bowl, out_min="f", out_eq=["FALSE"])

        # Test multiple restarts
        df_multi = gr.eval_min(
            md_bowl, out_min="f", out_geq=["g1"], out_leq=["g2"], n_restart=2,
        )
        self.assertTrue(df_multi.shape[0] == 2)


## Run tests
if __name__ == "__main__":
    unittest.main()



__all__ = [
    "plot_contour",
    "pt_contour",
    "plot_corrtile",
    "pt_corrtile",
    "plot_scattermat",
    "pt_scattermat",
    "plot_hists",
    "pt_hists",
    "plot_sinew_inputs",
    "pt_sinew_inputs",
    "plot_sinew_outputs",
    "pt_sinew_outputs",
    "plot_auto",
    "pt_auto",
    "plot_list",
    "set_uqtheme",
    "theme_grama",
    "theme_uqbook",

    "qual1",
    "qual2",
    "qual3",
    "qual4",
]

from grama import (
    add_pipe,
    pipe,
    case_when,
    Intention,
    tf_pivot_longer,
    tf_outer,
    tf_select,
    tf_rename,
    tf_filter,
    tf_mutate,
)
from .string_helpers import str_replace

from plotnine import aes, annotate, ggplot, facet_grid, facet_wrap, labs, guides
from plotnine import theme, theme_void, theme_minimal
from plotnine import element_blank, element_line, element_text, element_rect
from plotnine import geoms
from plotnine import (
    scale_x_continuous,
    scale_y_continuous,
    scale_fill_gradient,
    scale_fill_gradient2,
    scale_fill_gradientn,
)
from plotnine import (
    geom_blank,
    geom_point,
    geom_density,
    geom_histogram,
    geom_line,
    geom_tile,
    geom_text,
    geom_segment,
)

from matplotlib import gridspec
from pandas import melt

from toolz import curry

## Set color codes
##################################################
white = "#ffffff"
grey20 = "#2e2e2e"
grey50 = "#7d7d7d"
grey80 = "#cccccc"
black = "#000000"

# by David Nichols, https://davidmathlogic.com/colorblind/
qual1 = "#D81B60"
qual2 = "#1E88E5"
qual3 = "#FFC107"
qual4 = "#004D40"

## Helper functions
##################################################
def _sci_format(v):
    r"""Scientific format
    """
    return (
        ["{0:1.1e}".format(v[0])] +
        [""] * (len(v) - 2) +
        ["\n{0:1.1e}".format(v[-1])]
    )

## Function-specific plot functions
##################################################
## eval_contour
@curry
def plot_contour(df, var=None, out="out", level="level", aux=False, color="full"):
    r"""Plot 2d contours

    Plot contours. Usually called as a dispatch from plot_auto().

    Args:
        var (array of str): Variables for plot axes
        out (str): Name of output identifier column
        level (str): Name of level identifier column
        aux (bool): Auxillary variables present?
        color (str): Color mode; options: "full", "bw"

    Returns:
        ggplot: Contour image

    Examples::

        import grama as gr
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()
        (
            md_beam
            >> gr.ev_contour(
                var=["w", "t"],
                out=["g_stress"],
                # Set auxilliary inputs to nominal levels
                df=gr.eval_nominal(md_beam, df_det="nom"),
            )
            >> gr.pt_auto()
        )

    """
    # Check invariants
    if var is None:
        raise ValueError("Must provide input columns list as keyword var")
    if aux:
        raise ValueError(
            "Autoplot plot_contour not designed to handle auxiliary variables. " +
            "Regenerate contour data with fixed auxilary variables, " +
            "or try creating a manual plot."
        )

    if color == "full":
        return (
            df
            >> ggplot()
            + geom_segment(
                aes(
                    var[0],
                    var[1],
                    xend=var[0]+"_end",
                    yend=var[1]+"_end",
                    linetype=out,
                    color=level,
                )
            )
            + theme_grama()
        )
    elif color == "bw":
        return (
            df
            >> ggplot()
            + geom_segment(
                aes(
                    var[0],
                    var[1],
                    xend=var[0]+"_end",
                    yend=var[1]+"_end",
                    linetype=out,
                    group=level,
                )
            )
            + theme_grama()
        )
    else:
        raise ValueError("Color mode {} not recognized.".format(color))


pt_contour = add_pipe(plot_contour)

## tran_iocorr
# --------------------------------------------------
@curry
def plot_corrtile(df, var=None, out=None, corr=None, color="full"):
    r"""
    """
    if color == "full":
        return (
            df
            >> ggplot(aes(var, out))
            + geom_tile(aes(fill=corr))
            + scale_fill_gradient2(name="Corr", midpoint=0)
            + theme_grama()
            + theme(axis_text_x=element_text(angle=270))
        )
    elif color == "bw":
        DF = Intention()
        return (
            df
            >> tf_mutate(
                sign=case_when(
                    [DF[corr] > 0, "+"],
                    [DF[corr] < 0, "-"],
                    [True, None],
                )
            )
            >> ggplot(aes(var, out))
            + geom_tile(aes(fill=corr))
            + geom_text(aes(label="sign"), size=12, color="white")
            + scale_fill_gradientn(
                name="Corr",
                colors=("black", "white", "black"),
                values=(0, 0.5, 1),
            )
            + theme_grama()
            + theme(axis_text_x=element_text(angle=270))
        )
    else:
        raise ValueError("Color mode {} not recognized.".format(color))

pt_corrtile = add_pipe(plot_corrtile)

## tran_sobol
# --------------------------------------------------
@curry
def plot_sobol_outputs(df, idx=None, color="full"):
    r"""
    """
    df_data = (
        df
        >> tf_pivot_longer(
            columns=df.drop(idx, axis=1).columns,
            names_to="out",
            values_to="S"
        )
    )
    df_data[idx] = str_replace(df_data[idx], "^S_", "")

    if color == "full":
        return (
            df_data
            >> ggplot(aes(idx, "out"))
            + geom_tile(aes(fill="S"))
            + scale_fill_gradient(name="Sobol' Index", breaks=(0, 0.5, 1), limits=(0, 1))
            + theme_grama()
            + theme(axis_text_x=element_text(angle=270))
            + labs(
                x="var",
                y="out",
            )
        )
    elif color == "bw":
        return (
            df_data
            >> ggplot(aes(idx, "out"))
            + geom_tile(aes(fill="S"))
            + scale_fill_gradient(name="Sobol' Index", low="white", high="black", breaks=(0, 0.5, 1), limits=(0, 1))
            + theme_grama()
            + theme(axis_text_x=element_text(angle=270))
            + labs(
                x="var",
                y="out",
            )
        )
    else:
        raise ValueError("Color mode {} not recognized.".format(color))

pt_corrtile = add_pipe(plot_corrtile)

## Sample
# --------------------------------------------------
@curry
def plot_scattermat(df, var=None, color="full"):
    r"""Create a scatterplot matrix

    Create a scatterplot matrix. Often used to visualize a design (set of inputs
    points) before evaluating the functions.

    Usually called as a dispatch from plot_auto().

    Args:
        var (list of strings): Variables to plot
        color (str): Color mode; value ignored as default vis is black & white

    Returns:
        ggplot: Scatterplot matrix

    Examples::

        import grama as gr
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()
        ## Dispatch from autoplotter
        (
            md_beam
            >> gr.ev_sample(n=100, df_det="nom", skip=True)
            >> gr.pt_auto()
        )
        ## Re-create plot without metadata
        (
            md_beam
            >> gr.ev_sample(n=100, df_det="nom")
            >> gr.pt_scattermat(var=md.var)
        )

    """
    if var is None:
        raise ValueError("Must provide input columns list as keyword var")

    ## Define helpers
    labels_blank = lambda v: [""] * len(v)
    breaks_min = lambda lims: (lims[0], 0.5 * (lims[0] + lims[1]), lims[1])

    ## Make blank figure
    fig = (
        df
        >> ggplot()
        + geom_blank()
        + theme_void()
    ).draw(show=False)

    gs = gridspec.GridSpec(len(var), len(var))
    for i, v1 in enumerate(var):
        for j, v2 in enumerate(var):
            ax = fig.add_subplot(gs[i, j])
            ## Switch labels
            if j == 0:
                labels_y = _sci_format
            else:
                labels_y = labels_blank
            if i == len(var) - 1:
                labels_x = _sci_format
            else:
                labels_x = labels_blank

            ## Density
            if i == j:
                xmid = 0.5 * (
                    df[v1].min() + df[v1].max()
                )

                p = (
                    df
                    >> ggplot(aes(v1))
                    + geom_density()
                    + scale_x_continuous(
                        breaks=breaks_min,
                        labels=labels_x,
                    )
                    + scale_y_continuous(
                        breaks=breaks_min,
                        labels=labels_y,
                    )
                    + annotate(
                        "label",
                        x=xmid,
                        y=0,
                        label=v1,
                        va="bottom",
                    )
                    + theme_grama()
                    + labs(title=v1)
                )

            ## Scatterplot
            else:
                p = (
                    df
                    >> ggplot(aes(v2, v1))
                    + geom_point()
                    + scale_x_continuous(
                        breaks=breaks_min,
                        labels=labels_x,
                    )
                    + scale_y_continuous(
                        breaks=breaks_min,
                        labels=labels_y,
                    )
                    + theme_grama()
                    + theme(
                        axis_title=element_text(va="top", size=12),
                    )
                )

            _ = p._draw_using_figure(fig, [ax])


    ## Plot
    # NB Returning the figure causes a "double plot" in Jupyter....
    fig.show()


pt_scattermat = add_pipe(plot_scattermat)


@curry
def plot_hists(df, out=None, color="full", **kwargs):
    r"""Construct histograms

    Create a set of histograms. Often used to visualize the results of random
    sampling for multiple outputs.

    Usually called as a dispatch from plot_auto().

    Args:
        out (list of strings): Variables to plot
        color (str): Color mode; value ignored as default vis is black & white

    Returns:
        Seaborn histogram plot

    Examples::

        import grama as gr
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()
        ## Dispatch from autoplotter
        (
            md_beam
            >> gr.ev_sample(n=100, df_det="nom")
            >> gr.pt_auto()
        )
        ## Re-create without metadata
        (
            md_beam
            >> gr.ev_sample(n=100, df_det="nom")
            >> gr.pt_hists(out=md.out)
        )

    """
    if out is None:
        raise ValueError("Must provide input columns list as keyword out")

    return (
        df
        >> tf_pivot_longer(
            columns=out,
            names_to="var",
            values_to="value",
        )
        >> ggplot(aes("value"))
        + geom_histogram(bins=30)
        + facet_wrap("var", scales="free")
        + theme_grama()
        + theme(panel_spacing=0.40)
        + labs(
            x="Output Value",
            y="Count",
        )
    )


pt_hists = add_pipe(plot_hists)

## Sinew plots
# --------------------------------------------------
@curry
def plot_sinew_inputs(df, var=None, color="full", sweep_ind="sweep_ind"):
    r"""Inspect a sinew design

    Create a scatterplot matrix with hues. Often used to visualize a sinew
    design before evaluating the model functions.

    Usually called as a dispatch from plot_auto().

    Args:
        df (Pandas DataFrame): Input design data
        var (list of strings): Variables to plot
        sweep_ind (string): Sweep index column in df
        color (str): Color mode; options: "full", "bw"

    Returns:
        Seaborn scatterplot matrix

    Examples::

        import grama as gr
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()
        ## Dispatch from autoplotter
        (
            md_beam
            >> gr.ev_sinews(df_det="swp", skip=True)
            >> gr.pt_auto()
        )
        ## Re-create without metadata
        (
            md_beam
            >> gr.ev_sinews(df_det="swp")
            >> gr.pt_sinew_inputs(var=md.var)
        )

    """
    if var is None:
        raise ValueError("Must provide input columns list as keyword var")
    if not (color in ("full", "bw")):
        raise ValueError("Color mode {} not recognized.".format(color))

    ## Define helpers
    labels_blank = lambda v: [""] * len(v)
    breaks_min = lambda lims: (lims[0], 0.5 * (lims[0] + lims[1]), lims[1])

    ## Make blank figure
    fig = (
        df
        >> ggplot()
        + geom_blank()
        + theme_void()
    ).draw(show=False)

    gs = gridspec.GridSpec(len(var), len(var))
    for i, v1 in enumerate(var):
        for j, v2 in enumerate(var):
            ax = fig.add_subplot(gs[i, j])
            ## Switch labels
            if j == 0:
                labels_y = _sci_format
            else:
                labels_y = labels_blank
            if i == len(var) - 1:
                labels_x = _sci_format
            else:
                labels_x = labels_blank

            ## Label
            if i == j:
                p = (
                    df
                    >> ggplot()
                    + annotate(
                        "label",
                        x=0,
                        y=0,
                        label=v1,
                    )
                    + theme_void()
                    + guides(color=None, linetype=None)
                )

            ## Scatterplot
            else:
                if color == "full":
                    p = (
                        df
                        >> ggplot(aes(v2, v1, color="factor("+sweep_ind+")"))
                        + geom_point(size=0.1)
                        + scale_x_continuous(
                            breaks=breaks_min,
                            labels=labels_x,
                        )
                        + scale_y_continuous(
                            breaks=breaks_min,
                            labels=labels_y,
                        )
                        + guides(color=None)
                        + theme_grama()
                        + theme(
                            axis_title=element_text(va="top", size=12),
                        )
                    )
                elif color == "bw":
                    p = (
                        df
                        >> ggplot(aes(v2, v1))
                        + geom_point(size=0.1)
                        + scale_x_continuous(
                            breaks=breaks_min,
                            labels=labels_x,
                        )
                        + scale_y_continuous(
                            breaks=breaks_min,
                            labels=labels_y,
                        )
                        + guides(color=None)
                        + theme_grama()
                        + theme(
                            axis_title=element_text(va="top", size=12),
                        )
                    )

            _ = p._draw_using_figure(fig, [ax])


    ## Plot
    # NB Returning the figure causes a "double plot" in Jupyter....
    fig.show()


pt_sinew_inputs = add_pipe(plot_sinew_inputs)


@curry
def plot_sinew_outputs(
    df, var=None, out=None, sweep_ind="sweep_ind", sweep_var="sweep_var", color="full",
):
    r"""Construct sinew plot

    Create a relational lineplot with hues for each sweep. Often used to
    visualize the outputs of a sinew design.

    Usually called as a dispatch from plot_auto().

    Args:
        df (Pandas DataFrame): Input design data with output results
        var (list of strings): Variables to plot
        out (list of strings): Outputs to plot
        sweep_ind (string): Sweep index column in df
        sweep_var (string): Swept variable column in df
        color (str): Color mode; options: "full", "bw"

    Returns:
        Relational lineplot

    Examples::

        import grama as gr
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()
        ## Dispatch from autoplotter
        (
            md_beam
            >> gr.ev_sinews(df_det="swp")
            >> gr.pt_auto()
        )
        ## Re-create without metadata
        (
            md_beam
            >> gr.ev_sinews(df_det="swp")
            >> gr.pt_sinew_inputs(var=md.var, out=md.out)
        )

    """
    if var is None:
        raise ValueError("Must provide input columns list as keyword arg var")
    if out is None:
        raise ValueError("Must provide output columns list as keyword arg out")

    ## Prepare data
    # Gather inputs
    id_vars = [col for col in df.columns if col not in var]
    df_tmp = melt(df, id_vars, var, "_var", "_x")

    # Gather outputs
    id_vars = [col for col in df_tmp.columns if col not in out]
    df_plot = melt(df_tmp, id_vars, out, "_out", "_y")

    # Filter off-sweep values
    df_plot = df_plot[df_plot[sweep_var] == df_plot["_var"]]

    breaks_min = lambda lims: (lims[0], 0.5 * (lims[0] + lims[1]), lims[1])
    if color == "full":
        return (
            df_plot
            >> ggplot(aes(
                "_x",
                "_y",
                color="factor(" + sweep_ind + ")",
                group="factor(" + sweep_ind + ")",
            ))
            + geom_line()
            + facet_grid("_out~_var", scales="free")

            + scale_x_continuous(
                breaks=breaks_min,
                labels=_sci_format,
            )
            + scale_y_continuous(
                breaks=breaks_min,
                labels=_sci_format,
            )
            + guides(color=None)
            + theme_grama()
            + theme(
                strip_text_y=element_text(angle=270),
                panel_border=element_rect(color="black", size=0.5),
            )
            + labs(
                x="Input Value",
                y="Output Value",
            )
        )
    elif color == "bw":
        return (
            df_plot
            >> ggplot(aes(
                "_x",
                "_y",
                linetype="factor(" + sweep_ind + ")",
                group="factor(" + sweep_ind + ")",
            ))
            + geom_line()
            + facet_grid("_out~_var", scales="free")

            + scale_x_continuous(
                breaks=breaks_min,
                labels=_sci_format,
            )
            + scale_y_continuous(
                breaks=breaks_min,
                labels=_sci_format,
            )
            + guides(linetype=None)
            + theme_grama()
            + theme(
                strip_text_y=element_text(angle=0),
                panel_border=element_rect(color="black", size=0.5),
            )
            + labs(
                x="Input Value",
                y="Output Value",
            )
        )
    else:
        raise ValueError("Color mode {} not recognized.".format(color))


pt_sinew_outputs = add_pipe(plot_sinew_outputs)

## Autoplot dispatcher
## ##################################################
plot_list = {
    "contour": plot_contour,
    "sinew_inputs": plot_sinew_inputs,
    "sinew_outputs": plot_sinew_outputs,
    "sample_inputs": plot_scattermat,
    "sample_outputs": plot_hists,
    "iocorr": plot_corrtile,
    "sobol_outputs": plot_sobol_outputs,
}


@curry
def plot_auto(df, color="full"):
    r"""Automagic plotting

    Convenience tool for various grama outputs. Prints delegated plotting function, which can be called manually with different arguments for more tailored plotting.

    Args:
        df (DataFrame): Data output from appropriate grama routine. See gr.plot_list.keys() for list of supported methods.
        color (str): Color mode; options: "full", "bw"

    Returns:
        Plot results

    """
    try:
        d = df._plot_info
    except AttributeError:
        raise AttributeError(
            "'{}' object has no attribute _plot_info. Use plot_auto() for grama outputs only.".format(
                "DataFrame"
            )
        )

    try:
        plot_fcn = plot_list[d["type"]]
    except KeyError:
        raise KeyError("'{}' Plot type not implemented.".format(d["type"]))
    plt_kwargs = {key: value for key, value in d.items() if key != "type"}

    print("Calling {0:}....".format(plot_fcn.__name__))

    return plot_fcn(df, color=color, **plt_kwargs)


pt_auto = add_pipe(plot_auto)


## UQBook theme
## ##################################################
def set_uqtheme():
    """Override default geometry colors
    """

    geoms.geom_histogram.DEFAULT_AES['fill'] = grey50
    geoms.geom_histogram.DEFAULT_AES['color'] = black

def theme_grama():
    return (
        theme_minimal()
        + theme(
            axis_text=element_text(size=12),
            axis_title=element_text(size=14),
            strip_text=element_text(size=12),
            strip_text_y=element_text(size=12, vjust=0.75),
            strip_background=element_rect(color="black", fill="white", size=0.5),
            legend_title=element_text(size=12),
            legend_text=element_text(size=10),
        )
    )

def theme_uqbook():
    return (
        theme_grama()
        + theme(
            figure_size=(5, 3),
            legend_position="bottom",
            legend_box_spacing=0.5,
            legend_direction="horizontal",
        )
    )



import numpy as np
import pandas as pd
from scipy.stats import norm
import unittest

from context import grama as gr
from context import data

X = gr.Intention()

##==============================================================================
## filter helper tests
##==============================================================================


class TestHelpers(unittest.TestCase):
    def test_var_in(self):
        df = data.df_diamonds[["cut"]].head(10)
        d1 = df[(df.cut == "Ideal") | (df.cut == "Premium")].reset_index(drop=True)
        d2 = df >> gr.tf_filter(gr.var_in(X.cut, ["Ideal", "Premium"]))

        self.assertTrue(d1.equals(d2))



import numpy as np
import pandas as pd
import unittest

from context import grama as gr
from context import data

X = gr.Intention()

##==============================================================================
## transform test functions
##==============================================================================
def group_mutate_helper(df):
    df["testcol"] = df["x"] * df.shape[0]
    return df


class testTransform(unittest.TestCase):
    def test_mutate(self):
        df = data.df_diamonds.copy()
        df["testcol"] = 1
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_mutate(testcol=1)))
        df["testcol"] = df["x"]
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_mutate(testcol=X.x)))
        df["testcol"] = df["x"] * df["y"]
        self.assertTrue(df.equals(data.df_diamonds >> gr.tf_mutate(testcol=X.x * X.y)))
        df["testcol"] = df["x"].mean()
        self.assertTrue(
            df.equals(data.df_diamonds >> gr.tf_mutate(testcol=np.mean(X.x)))
        )

    def test_group_mutate(self):
        df = data.df_diamonds.copy()
        df = df.groupby("cut").apply(group_mutate_helper)
        d = (
            data.df_diamonds
            >> gr.tf_group_by("cut")
            >> gr.tf_mutate(testcol=X.x * X.shape[0])
            >> gr.tf_ungroup()
        )
        self.assertTrue(df.equals(d.sort_index()))

    def test_mutate_if(self):
        df = data.df_diamonds.copy()
        for col in df:
            try:
                if max(df[col]) < 10:
                    df[col] *= 2
            except:
                pass
        self.assertTrue(
            df.equals(
                data.df_diamonds
                >> gr.tf_mutate_if(lambda col: max(col) < 10, lambda row: row * 2)
            )
        )
        df = data.df_diamonds.copy()
        for col in df:
            try:
                if any(df[col].str.contains(".")):
                    df[col] = df[col].str.lower()
            except:
                pass
        self.assertTrue(
            df.equals(
                data.df_diamonds
                >> gr.tf_mutate_if(
                    lambda col: any(col.str.contains(".")), lambda row: row.str.lower()
                )
            )
        )
        df = data.df_diamonds.copy()
        for col in df:
            try:
                if min(df[col]) < 1 and mean(df[col]) < 4:
                    df[col] *= -1
            except:
                pass
        self.assertTrue(
            df.equals(
                data.df_diamonds
                >> gr.tf_mutate_if(
                    lambda col: min(col) < 1 and mean(col) < 4, lambda row: -row
                )
            )
        )



import numpy as np
import pandas as pd
import unittest
import io
import sys

from context import grama as gr
from context import data

## Test the built-in datasets
##################################################
class TestData(unittest.TestCase):

    def setUp(self):
        pass

    def test_make(self):
        df_stang = data.df_stang

    def test_install(self):
        # Only works if grama installed locally!
        from grama.data import df_stang



__all__ = [
    "var_in",
    "is_nan",
    "not_nan",
]

from .base import make_symbolic
from numpy import isnan, logical_not
from pandas import Series


# ------------------------------------------------------------------------------
# Mask helpers
# ------------------------------------------------------------------------------


@make_symbolic
def var_in(series, collection):
    """Determine if value is in collection

    Returns a boolean series where each entry denotes inclusion in the provided collection. Intended for use in tran_filter() calls.

    Args:
        series: column to compute inclusion bools
        collection: set for inclusion calcs

    """
    bools = Series([s in collection for s in series])

    return bools


@make_symbolic
def is_nan(series, inv=False):
    """Determine if nan

    Returns a boolean series where each entry denotes NaN or not. Intended for use in tran_filter() calls.

    Args:
        series (Pandas series): column to compute NaN bools
        inv (bool): Invert logic

    """
    bools = isnan(series)

    if inv:
        return logical_not(bools)
    return bools


@make_symbolic
def not_nan(series, inv=False):
    """Determine if NOT nan

    Returns a boolean series where each entry denotes NOT NaN or yes. Intended for use in tran_filter() calls.

    Args:
        series (Pandas series): column to compute NOT NaN bools
        inv (bool): Invert logic

    """
    bools = isnan(series)

    if inv:
        return bools
    return logical_not(bools)



__all__ = [
    "marg_gkde",
    "marg_fit",
    "marg_mom",
    "marg_trunc",
    "Marginal",
    "MarginalNamed",
    "MarginalGKDE",
    "MarginalTruncated",
    "param_dist",
    "valid_dist",
]

import copy
import warnings
from grama import make_symbolic
from abc import ABC, abstractmethod
from numpy import zeros, array, Inf, concatenate, sqrt
from numpy import min as npmin
from numpy import max as npmax
from numpy.random import uniform as runif
from pandas import DataFrame
from scipy.optimize import root_scalar, root
from scipy.stats import alpha, anglit, arcsine, argus, beta, betaprime, \
    bradford, burr, burr12, cauchy, chi, chi2, cosine, crystalball, dgamma, \
    dweibull, erlang, expon, exponnorm, exponweib, exponpow, f, fatiguelife, \
    fisk, foldcauchy, foldnorm, gaussian_kde, genlogistic, gennorm, genpareto, \
    genexpon, genextreme, gausshyper, gamma, gengamma, genhalflogistic, \
    gibrat, gompertz, gumbel_r, gumbel_l, halfcauchy, halflogistic, \
    halfnorm, halfgennorm, hypsecant, invgamma, invgauss, invweibull, \
    johnsonsb, johnsonsu, kappa4, kappa3, ksone, kstwobign, laplace, levy, \
    levy_l, levy_stable, logistic, loggamma, loglaplace, lognorm, lomax, \
    maxwell, mielke, moyal, nakagami, ncx2, ncf, nct, norm, norminvgauss, \
    pareto, pearson3, powerlaw, powerlognorm, powernorm, rdist, rayleigh, \
    rice, recipinvgauss, skewnorm, t, trapz, triang, truncexpon, truncnorm, \
    tukeylambda, uniform, vonmises, vonmises_line, wald, weibull_min, \
    weibull_max, wrapcauchy


## Scipy metadata
valid_dist = {
    "alpha": alpha,
    "anglit": anglit,
    "arcsine": arcsine,
    "argus": argus,
    "beta": beta,
    "betaprime": betaprime,
    "bradford": bradford,
    "burr": burr,
    "burr12": burr12,
    "cauchy": cauchy,
    "chi": chi,
    "chi2": chi2,
    "cosine": cosine,
    "crystalball": crystalball,
    "dgamma": dgamma,
    "dweibull": dweibull,
    "erlang": erlang,
    "expon": expon,
    "exponnorm": exponnorm,
    "exponweib": exponweib,
    "exponpow": exponpow,
    "f": f,
    "fatiguelife": fatiguelife,
    "fisk": fisk,
    "foldcauchy": foldcauchy,
    "foldnorm": foldnorm,
    # "frechet_r": frechet_r,
    # "frechet_l": frechet_l,
    "genlogistic": genlogistic,
    "gennorm": gennorm,
    "genpareto": genpareto,
    "genexpon": genexpon,
    "genextreme": genextreme,
    "gausshyper": gausshyper,
    "gamma": gamma,
    "gengamma": gengamma,
    "genhalflogistic": genhalflogistic,
    # "geninvgauss": geninvgauss,
    "gibrat": gibrat,
    "gompertz": gompertz,
    "gumbel_r": gumbel_r,
    "gumbel_l": gumbel_l,
    "halfcauchy": halfcauchy,
    "halflogistic": halflogistic,
    "halfnorm": halfnorm,
    "halfgennorm": halfgennorm,
    "hypsecant": hypsecant,
    "invgamma": invgamma,
    "invgauss": invgauss,
    "invweibull": invweibull,
    "johnsonsb": johnsonsb,
    "johnsonsu": johnsonsu,
    "kappa4": kappa4,
    "kappa3": kappa3,
    "ksone": ksone,
    "kstwobign": kstwobign,
    "laplace": laplace,
    "levy": levy,
    "levy_l": levy_l,
    "levy_stable": levy_stable,
    "logistic": logistic,
    "loggamma": loggamma,
    "loglaplace": loglaplace,
    "lognorm": lognorm,
    # "loguniform": loguniform,
    "lomax": lomax,
    "maxwell": maxwell,
    "mielke": mielke,
    "moyal": moyal,
    "nakagami": nakagami,
    "ncx2": ncx2,
    "ncf": ncf,
    "nct": nct,
    "norm": norm,
    "norminvgauss": norminvgauss,
    "pareto": pareto,
    "pearson3": pearson3,
    "powerlaw": powerlaw,
    "powerlognorm": powerlognorm,
    "powernorm": powernorm,
    "rdist": rdist,
    "rayleigh": rayleigh,
    "rice": rice,
    "recipinvgauss": recipinvgauss,
    "skewnorm": skewnorm,
    "t": t,
    "trapz": trapz,
    "triang": triang,
    "truncexpon": truncexpon,
    "truncnorm": truncnorm,
    "tukeylambda": tukeylambda,
    "uniform": uniform,
    "vonmises": vonmises,
    "vonmises_line": vonmises_line,
    "wald": wald,
    "weibull_min": weibull_min,
    "weibull_max": weibull_max,
    "wrapcauchy": wrapcauchy,
}

param_dist = {
    "alpha": ["a", "loc", "scale"],
    "anglit": ["loc", "scale"],
    "arcsine": ["loc", "scale"],
    "argus": ["chi", "loc", "scale"],
    "beta": ["a", "b", "loc", "scale"],
    "betaprime": ["a", "b", "loc", "scale"],
    "bradford": ["c", "loc", "scale"],
    "burr": ["c", "d", "loc", "scale"],
    "burr12": ["c", "d", "loc", "scale"],
    "cauchy": ["loc", "scale"],
    "chi": ["df", "loc", "scale"],
    "chi2": ["df", "loc", "scale"],
    "cosine": ["loc", "scale"],
    "crystalball": ["beta", "m", "loc", "scale"],
    "dgamma": ["a", "loc", "scale"],
    "dweibull": ["c", "loc", "scale"],
    "erlang": ["loc", "scale"],
    "expon": ["loc", "scale"],
    "exponnorm": ["K", "loc", "scale"],
    "exponweib": ["a", "c", "loc", "scale"],
    "exponpow": ["b", "loc", "scale"],
    "f": ["dfn", "dfd", "loc", "scale"],
    "fatiguelife": ["c", "loc", "scale"],
    "fisk": ["c", "loc", "scale"],
    "foldcauchy": ["c", "loc", "scale"],
    "foldnorm": ["c", "loc", "scale"],
    "frechet_r": ["c", "loc", "scale"],
    "frechet_l": ["c", "loc", "scale"],
    "genlogistic": ["c", "loc", "scale"],
    "gennorm": ["beta", "loc", "scale"],
    "genpareto": ["c", "loc", "scale"],
    "genexpon": ["a", "b", "c", "loc", "scale"],
    "genextreme": ["c", "loc", "scale"],
    "gausshyper": ["a", "b", "c", "z", "loc", "scale"],
    "gamma": ["a", "loc", "scale"],
    "gengamma": ["a", "c", "loc", "scale"],
    "genhalflogistic": ["c", "loc", "scale"],
    "geninvgauss": ["p", "b", "loc", "scale"],
    "gibrat": ["loc", "scale"],
    "gompertz": ["c", "loc", "scale"],
    "gumbel_r": ["loc", "scale"],
    "gumbel_l": ["loc", "scale"],
    "halfcauchy": ["loc", "scale"],
    "halflogistic": ["loc", "scale"],
    "halfnorm": ["loc", "scale"],
    "halfgennorm": ["beta", "loc", "scale"],
    "hypsecant": ["loc", "scale"],
    "invgamma": ["a", "loc", "scale"],
    "invgauss": ["mu", "loc", "scale"],
    "invweibull": ["c", "loc", "scale"],
    "johnsonsb": ["a", "b", "loc", "scale"],
    "johnsonsu": ["a", "b", "loc", "scale"],
    "kappa4": ["h", "k", "loc", "scale"],
    "kappa3": ["a", "loc", "scale"],
    "ksone": ["n", "loc", "scale"],
    "kstwobign": ["loc", "scale"],
    "laplace": ["loc", "scale"],
    "levy": ["loc", "scale"],
    "levy_l": ["loc", "scale"],
    "levy_stable": ["alpha", "beta", "loc", "scale"],
    "logistic": ["loc", "scale"],
    "loggamma": ["c", "loc", "scale"],
    "loglaplace": ["c", "loc", "scale"],
    "lognorm": ["s", "loc", "scale"],
    "loguniform": ["a", "b", "loc", "scale"],
    "lomax": ["c", "loc", "scale"],
    "maxwell": ["loc", "scale"],
    "mielke": ["k", "s", "loc", "scale"],
    "moyal": ["loc", "scale"],
    "nakagami": ["nu", "loc", "scale"],
    "ncx2": ["df", "nc", "loc", "scale"],
    "ncf": ["dfn", "dfd", "nc", "loc", "scale"],
    "nct": ["df", "nc", "loc", "scale"],
    "norm": ["loc", "scale"],
    "norminvgauss": ["a", "b", "loc", "scale"],
    "pareto": ["b", "loc", "scale"],
    "pearson3": ["skew", "loc", "scale"],
    "powerlaw": ["a", "loc", "scale"],
    "powerlognorm": ["c", "s", "loc", "scale"],
    "powernorm": ["c", "loc", "scale"],
    "rdist": ["c", "loc", "scale"],
    "rayleigh": ["loc", "scale"],
    "rice": ["b", "loc", "scale"],
    "recipinvgauss": ["mu", "loc", "scale"],
    "semicircular": ["loc", "scale"],
    "skewnorm": ["a", "loc", "scale"],
    "t": ["df", "loc", "scale"],
    "trapz": ["c", "d", "loc", "scale"],
    "triang": ["c", "loc", "scale"],
    "truncexpon": ["b", "loc", "scale"],
    "truncnorm": ["a", "b", "loc", "scale"],
    "tukeylambda": ["lam", "loc", "scale"],
    "uniform": ["loc", "scale"],
    "vonmises": ["kappa", "loc", "scale"],
    "vonmises_line": ["kappa", "loc", "scale"],
    "wald": ["loc", "scale"],
    "weibull_min": ["c", "loc", "scale"],
    "weibull_max": ["c", "loc", "scale"],
    "wrapcauchy": ["c", "loc", "scale"],
}

## Marginal classes
##################################################
# Marginal parent class
class Marginal(ABC):
    """Parent class for marginal distributions
    """

    def __init__(self, sign=0):
        self.sign = sign

    @abstractmethod
    def copy(self):
        pass

    ## Fitting function
    @abstractmethod
    def fit(self, data):
        pass

    ## Density function
    @abstractmethod
    def d(self, x):
        pass

    ## Cumulative density function
    @abstractmethod
    def p(self, x):
        pass

    ## Quantile function
    @abstractmethod
    def q(self, p):
        pass

    ## Random variable sample
    def r(self, n):
        U = runif(size=n)
        return self.q(U)

    ## Summary
    @abstractmethod
    def summary(self):
        pass

    def __str__(self):
        return self.summary()

    def __repr__(self):
        return self.summary()


## Named marginal class
class MarginalNamed(Marginal):
    """Marginal using a named distribution from gr.valid_dist"""

    def __init__(self, d_name=None, d_param=None, **kw):
        super().__init__(**kw)

        self.d_name = d_name
        self.d_param = d_param

    def copy(self):
        new_marginal = MarginalNamed(
            sign=self.sign, d_name=self.d_name, d_param=copy.deepcopy(self.d_param)
        )

        return new_marginal

    ## Fitting function
    def fit(self, data, **kwargs):
        param = valid_dist[self.d_name].fit(data, **kwargs)
        self.d_param = dict(zip(param_dist[dist], param))

    ## Density function
    @make_symbolic
    def d(self, x):
        return valid_dist[self.d_name].pdf(x, **self.d_param)

    ## Cumulative density function
    @make_symbolic
    def p(self, x):
        return valid_dist[self.d_name].cdf(x, **self.d_param)

    ## Quantile function
    @make_symbolic
    def q(self, p):
        return valid_dist[self.d_name].ppf(p, **self.d_param)

    ## Summary
    def summary(self, dig=2):
        stats = valid_dist[self.d_name](**self.d_param).stats("mvsk")
        param = {
            "mean": "{0:4.3e}".format(stats[0].round(dig)),
            "s.d.": "{0:4.3e}".format(sqrt(stats[1]).round(dig)),
            "COV": round(sqrt(stats[1]) / stats[0], dig),
            "skew.": stats[2].round(dig),
            "kurt.": stats[3].round(dig) + 3, # full kurtosis
        }
        return "({0:+}) {1:}, {2:}".format(self.sign, self.d_name, param)


## Gaussian KDE marginal class
class MarginalGKDE(Marginal):
    """Marginal using scipy.stats.gaussian_kde"""

    def __init__(self, kde, atol=1e-6, **kw):
        super().__init__(**kw)

        self.kde = kde
        self.atol = atol
        self._set_bracket()

    def copy(self):
        new_marginal = MarginalGKDE(kde=copy.deepcopy(self.kde), atol=self.atol,)

        return new_marginal

    def _set_bracket(self):
        ## Calibrate the quantile brackets based on desired accuracy
        bracket = [npmin(self.kde.dataset), npmax(self.kde.dataset)]

        sol_lo = root_scalar(
            lambda x: self.atol - self.p(x),
            x0=bracket[0],
            x1=bracket[1],
            method="secant",
        )
        sol_hi = root_scalar(
            lambda x: 1 - self.atol - self.p(x),
            x0=bracket[0],
            x1=bracket[1],
            method="secant",
        )

        self.bracket = [sol_lo.root, sol_hi.root]

    ## Fitting function
    def fit(self, data):
        self.kde = gaussian_kde(data)
        self._set_bracket()

    ## Density function
    @make_symbolic
    def d(self, x):
        return self.kde.pdf(x)

    ## Cumulative density function
    @make_symbolic
    def p(self, x):
        try:
            return array([self.kde.integrate_box_1d(-Inf, v) for v in x])
        except TypeError:
            return self.kde.integrate_box_1d(-Inf, x)

    ## Quantile function
    @make_symbolic
    def q(self, p):
        p_bnd = self.p(self.bracket)

        ## Create scalar solver
        def qscalar(val):
            if val <= p_bnd[0]:
                return self.bracket[0]
            if val >= p_bnd[1]:
                return self.bracket[1]
            sol = root_scalar(
                lambda x: val - self.p(x),
                bracket=self.bracket,
                method="bisect",
                xtol=self.atol,
            )
            return sol.root

        ## Iterate over all given values
        try:
            res = zeros(len(p))
            for i in range(len(p)):
                res[i] = qscalar(p[i])
        except TypeError:
            res = qscalar(p)

        return res

    ## Summary
    def summary(self):
        p_bnd = self.p(self.bracket)

        return "({0:+}) gaussian KDE, n={1:2.1f}/{2:}, f={3:2.1e}, ".format(
            self.sign, self.kde.neff, self.kde.dataset.shape[1], self.kde.factor
        ) + "b=[{0:2.1e}, {1:2.1e}], a={2:1.0e}".format(
            self.bracket[0], self.bracket[1], self.atol
        )

## Named marginal class
class MarginalTruncated(Marginal):
    """Marginal a truncated Marginal"""

    def __init__(self, marg, lo, up, **kw):
        r"""
        Args:
            marg (gr.Marginal): Base Marginal to truncate
            lo (float): lower bound (a)
            up (float): upper bound (b)

        Preconditions:
            at least one of `lo` and `up` must not be None
        """
        super().__init__(**kw)

        self.marg = marg
        self.lo = lo
        self.up = up
        self.G_lo, self.G_up = self.truncate()

        if self.lo is None:
            self.lo = -Inf
        if self.up is None:
            self.up = +Inf

    def truncate(self):
        if self.lo is None:
            G_lo = 0
        else:
            G_lo = self.marg.p(self.lo)

        if self.up is None:
            G_up = 1
        else:
            G_up = self.marg.p(self.up)

        return G_lo, G_up

    def copy(self):
        new_marginal = MarginalTruncated(
            self.marg.copy(), self.lo, self.up, sign=self.sign,
        )

        return new_marginal

    ## Fitting function
    def fit(self, data, lo, up, **kwargs):
        # Re-fit the base marginal
        self.marg.fit(data, **kwargs)
        # Re-truncate the marginal
        self.lo = lo
        self.up = up
        self.G_lo, self.G_up = self.truncate()

    ## Density function
    @make_symbolic
    def d(self, x):
        return (
            self.marg.d(x) / (self.G_up - self.G_lo)
            # Zero-out values outside truncation
            *(self.lo <= x) * (x <= self.up)
        )

    ## Cumulative density function
    @make_symbolic
    def p(self, x):
        # Piecewise definition
        return (
            # Below lower bound
            + 0 * (x < self.lo)
            # Within bounds
            + (self.marg.p(x) - self.G_lo) / (self.G_up - self.G_lo)
              *(self.lo <= x) * (x <= self.up)
            # Above upper bound
            + 1 * (self.up < x)
        )

    ## Quantile function
    @make_symbolic
    def q(self, p):
        return self.marg.q( p * (self.G_up - self.G_lo) + self.G_lo )

    ## Summary
    def summary(self, dig=2):
        s_base = self.marg.summary()
        if (self.lo is not None) and (self.up is not None):
            s_trunc = "truncated to [{0:}, {1:}]".format(round(self.lo, dig), round(self.up, dig))
        elif self.lo is not None:
            s_trunc = "truncated to [{0:}, +\infty)".format(round(self.lo, dig))
        elif self.up is not None:
            s_trunc = "truncated to (+\infty, {0:}]".format(round(self.up, dig))

        return s_base + " " + s_trunc


## Marginal functions
##################################################
def marg_mom(
        dist,
        mean=None,
        sd=None,
        cov=None,
        var=None,
        skew=None,
        kurt=None,
        kurt_excess=None,
        lo=None,
        up=None,
        floc=None,
        sign=0,
        dict_x0=None,
):
    r"""Fit scipy.stats continuous distribution via moments

    Fit a continuous distribution using the method of moments. Select a distribution shape and provide numerical values for a convenient set of common moments.

    This routine uses a vector-output root finding routine to match the moments. You may set an optional initial guess for the distribution parameters using the dict_x0 argument.

    Args:
        dist (str): Name of distribution to fit

    Kwargs:
        mean (float): Mean of distribution
        sd (float): Standard deviation of distribution
        cov (float): Coefficient of Variation of distribution (sd / mean)
        var (float): Variance of distribution; only one of `sd` and `var` can be provided.
        skew (float): Skewness of distribution
        kurt (float): Kurtosis of distribution
        kurt_excess (float): Excess kurtosis of distribution; kurt_excess = kurt - 3.
            Only one of `kurt` and `kurt_excess` can be provided.
        lo (float): Lower bound, dist = "uniform" or "beta" only
        up (float): Upper bound, dist = "uniform" or "beta" only

        floc (float or None): Frozen value for location parameter
            Note that for distributions such as "lognorm" or "weibull_min",
            setting floc=0 selects the 2-parameter version of the distribution.

        sign (-1, 0, +1): Sign
        dict_x0 (dict): Dictionary of initial parameter guesses

    Returns:
        gr.MarginalNamed: Distribution

    Examples::

        import grama as gr
        ## Fit a normal distribution
        mg_norm = gr.marg_mom("norm", mean=0, sd=1)
        ## Fit a (3-parameter) lognormal distribution
        mg_lognorm = gr.marg_mom("lognorm", mean=1, sd=1, skew=1)
        ## Fit a lognormal, controlling kurtosis instead
        mg_lognorm = gr.marg_mom("lognorm", mean=1, sd=1, kurt=1)
        ## Fit a 2-parameter lognormal; no skewness or kurtosis needed
        mg_lognorm = gr.marg_mom("lognorm", mean=1, sd=1, floc=0)

        ## Not all moment combinations are feasible; this will fail
        gr.marg_mom("beta", mean=1, sd=1, skew=0, kurt=4)
        ## Skewness and kurtosis are related for the beta distribution;
        ## a different combination is feasible
        gr.marg_mom("beta", mean=1, sd=1, skew=0, kurt=2)

    """
    ## Number of distribution parameters
    n_param = len(param_dist[dist])
    if n_param > 4:
        raise NotImplementedError(
            "marg_nom does not yet handle distributions with more than 4 parameters"
        )
    if floc is not None:
        n_param = n_param - 1

    ## Handle bounded cases
    if (lo is not None) or (up is not None):
        # We can handle the uniform case without optimization
        if dist == "uniform":
            return MarginalNamed(
                sign=sign,
                d_name=dist,
                d_param={"loc": lo, "scale": up - lo},
            )
        elif dist == "beta":
            bounded = True
            n_param = n_param - 2
        else:
            raise ValueError("lo and up arguments can only be used with dist=='uniform' or 'beta'")
    else:
        bounded = False

    ## Check invariants
    if mean is None:
        raise ValueError("Must provide `mean` argument.")
    if (sd is None) and (var is None) and (cov is None):
        raise ValueError(
            "One of `sd`, `cov`, or `var` must be provided."
        )
    if sum([(not sd is None), (not var is None), (not cov is None)]) > 1:
        raise ValueError(
            "Only one of `sd`, `cov`, and `var` may be provided."
        )
    if (not kurt is None) and (not kurt_excess is None):
        raise ValueError(
            "Only one of `kurt` and `kurt_excess` may be provided."
        )

    ## Process arguments
    # Transform to "standard" moments
    if (not sd is None):
        var = sd**2
    if (not cov is None):
        var = (mean * cov)**2
    if (not kurt is None):
        kurt_excess = kurt - 3

    # Build up target moments
    s = "mv"
    m_target = array([mean, var])

    if (not skew is None):
        s = s + "s"
        m_target = concatenate((m_target, array([skew])))
    if (not kurt is None):
        s = s + "k"
        m_target = concatenate((m_target, array([kurt_excess])))
    n_provided = len(s)

    if n_provided < n_param:
        raise ValueError(
            "Insufficient moments provided; you must provide {} more moment(s).".format(
                n_param - n_provided
            )
        )
    if n_provided > n_param:
        raise ValueError(
            "Overdetermined; you must provide {} fewer moment(s).".format(
                n_provided - n_param
            )
        )

    ## Generate helper function for optimization
    # Get parameters to treat as variables for rootfinding
    if floc is not None:
        key_wk = {key for key in param_dist[dist] if key != "loc"}
    elif bounded:
        key_wk = {key for key in param_dist[dist] if (key != "loc" and key != "scale")}
    else:
        key_wk = copy.copy(param_dist[dist])
    # Generate a helper with the variable parameters and target moments
    if floc is not None:
        def _obj(v):
            kw = dict(zip(key_wk, v))
            kw["loc"] = floc
            return array(valid_dist[dist](**kw).stats(s)) - m_target
    elif bounded:
        def _obj(v):
            kw = dict(zip(key_wk, v))
            kw["loc"] = lo
            kw["scale"] = up - lo
            return array(valid_dist[dist](**kw).stats(s)) - m_target
    else:
        def _obj(v):
            kw = dict(zip(key_wk, v))
            return array(valid_dist[dist](**kw).stats(s)) - m_target


    ## Generate initial guess
    if dict_x0 is None:
        if dist == "lognorm":
            dict_x0 = dict(
                loc=0,
                scale=mean,
                s=sqrt(var) / mean,
            )

        elif dist == "weibull_min":
            dict_x0 = dict(
                loc=0,
                scale=mean,
                c=1,
            )

        elif dist == "beta":
            dict_x0 = dict(
                loc=mean - sqrt(var),
                scale=sqrt(var),
                a=1 / var / 8,
                b=1 / var / 8,
            )

        else:
            # General-purpose initial guesses
            dict_x0 = dict(
                loc=mean,
                scale=sqrt(var),
                a=1,
                b=1,
                s=1,
                df=10,
                c=10,
                beta=1,
                #K=None,
                #chi=None,
            )

    # Repackage for optimizer
    x0 = array([dict_x0[key] for key in key_wk])

    ## Run multidimensional root finding
    res = root(_obj, x0)

    ## Check for failed optimization
    if res.success is False:
        raise RuntimeError(
            "Moment matching failed; initial guess may be poor, or requested "
            "moments may be infeasible. Try setting `dict_x0`. " +
            "Printing optimization results for debugging:\n\n{}".format(res)
        )

    ## Repackage and return
    param = dict(zip(key_wk, res.x))
    if floc is not None:
        param["loc"] = floc
    elif bounded:
        param["loc"] = lo
        param["scale"] = up - lo
    return MarginalNamed(sign=sign, d_name=dist, d_param=param)

## Fit a named scipy.stats distribution
def marg_fit(dist, data, name=True, sign=None, **kwargs):
    r"""Fit scipy.stats continuous distirbution

    Fits a scipy.stats continuous distribution. Intended to be used to define a marginal distribution from data.

    Arguments:
        dist (str): Distribution to fit
        data (iterable): Data for fit
        name (bool): Include distribution name?
        sign (bool): Include sign? (Optional)

        loc (float): Initial guess for location `loc` parameter (Optional)
        scale (float): Initial guess for scale `scale` parameter (Optional)

        floc (float): Value to fix the location `loc` parameter (Optional)
            Note that for distributions such as "lognorm" or "weibull_min",
            setting floc=0 selects the 2-parameter version of the distribution.
        fscale (float): Value to fix the location `scale` parameter (Optional)
        f* (float): Value to fix the specified shape parameter (Optional)
            e.g. give fc to fix the `c` parameter

    Returns:
        gr.MarginalNamed: Distribution

    Examples::

        import grama as gr
        from grama.data import df_shewhart
        # Fit normal distribution
        mg_normal = gr.marg_named(
            "norm",
            df_shewhart.tensile_strength,
        )
        # Fit two-parameter Weibull distribution
        mg_weibull2 = gr.marg_named(
            "weibull_min",
            df_shewhart.tensile_strength,
            floc=0,        # 2-parameter has frozen loc == 0
        )
        # Fit three-parameter Weibull distribution
        mg_weibull3 = gr.marg_named(
            "weibull_min",
            df_shewhart.tensile_strength,
            loc=0,        # 3-parameter fit tends to be unstable;
                          # an inital guess helps stabilize fit
        )
        # Inspect fits with QQ plot
        (
            df_shewhart
            >> gr.tf_mutate(
                q_normal=gr.qqvals(DF.tensile_strength, marg=mg_normal),
                q_weibull2=gr.qqvals(DF.tensile_strength, marg=mg_weibull2),
            )
            >> gr.tf_pivot_longer(
                columns=[
                    "q_normal",
                    "q_weibull2",
                ],
                names_to=[".value", "Distribution"],
                names_sep="_"
            )

            >> gr.ggplot(gr.aes("q", "tensile_strength"))
            + gr.geom_abline(intercept=0, slope=1, linetype="dashed")
            + gr.geom_point(gr.aes(color="Distribution"))
        )

    """
    ## Catch case where user provides entire DataFrame
    if isinstance(data, DataFrame):
        raise ValueError("`data` argument must be a single column; try data.var")

    ## Fit the distribution
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        param = valid_dist[dist].fit(data, **kwargs)

    param = dict(zip(param_dist[dist], param))

    if sign is not None:
        if not (sign in [-1, 0, +1]):
            raise ValueError("Invalid `sign`")
    else:
        sign = 0

    return MarginalNamed(sign=sign, d_name=dist, d_param=param)


## Fit a gaussian kernel density estimate (KDE) to data
def marg_gkde(data, sign=None):
    r"""Fit a gaussian KDE to data

    Fits a gaussian kernel density estimate (KDE) to data.

    Args:
        data (iterable): Data for fit
        sign (bool): Include sign? (Optional)

    Returns:
        gr.MarginalGKDE: Marginal distribution

    Examples::

        import grama as gr
        from grama.data import df_stang
        md = (
            gr.Model("Marginal Example")
            >> gr.cp_marginals(
                E=gr.marg_gkde(df_stang.E),
                mu=gr.marg_gkde(df_stang.mu),
            )
        )
        md

    """
    ## Catch case where user provides entire DataFrame
    if isinstance(data, DataFrame):
        raise ValueError("`data` argument must be a single column; try data.var")

    kde = gaussian_kde(data)
    if sign is not None:
        if not (sign in [-1, 0, +1]):
            raise ValueError("Invalid `sign`")
    else:
        sign = 0

    return MarginalGKDE(kde, sign=sign)

## Truncate an existing marginal
def marg_trunc(marg, lo=None, up=None, **kwargs):
    r"""Truncate an existing Marginal()

    Truncate

    Args:
        marg (gr.Marginal): Marginal to truncate

    Kwargs:
        lo (float or None): Lower bound for truncation; None indicates no lower bound (-\infty)
        up (float or None): Upper bound for truncation; None indicates no upper bound (+\infty)

    Returns:
        gr.MarginalTruncated: Truncated distribution

    Examples::

        import grama as gr
        ## Fit a normal distribution
        mg_norm = gr.marg_mom("norm", mean=0, sd=1)
        ## Truncate the distribution
        mg_norm_t1 = gr.marg_trunc(mg_norm, lo=-1, up=+1)
        ## Truncate the lower-end of the distribution only
        mg_norm_t2 = gr.marg_trunc(mg_norm, lo=0)
    """
    # Make a working copy
    mg_wk = marg.copy()

    # Check invariants
    if (lo is None) and (up is None):
        raise ValueError("Cannot truncate without at least one of `lo` or `up`.")

    # Construct the marginal
    mg = MarginalTruncated(mg_wk, lo, up, **kwargs)

    return mg



__all__ = [
    "tran_polyridge",
    "tf_polyridge",
    "fit_polyridge",
    "ft_polyridge",
]

import copy
from grama import add_pipe, Function, Model
from numpy import number as npnumber
from numpy.random import seed as npseed
from pandas import DataFrame
from toolz import curry

from .polyridge import PolynomialRidgeApproximation

## Helper functions
# --------------------------------------------------
class FunctionPoly(Function):
    def __init__(self, regressor, var, out, name, runtime):
        """

        Args:
            regressor (scikit-style Regressor):
        """
        self.regressor = regressor
        self.var = var
        self.out = list(map(lambda s: s + "_mean", out))
        self.name = name
        self.runtime = runtime

    def copy(self):
        """Make a copy"""
        func_new = FunctionPoly(
            copy.deepcopy(self.regressor),
            copy.deepcopy(self.var),
            copy.deepcopy(self.out),
            copy.deepcopy(self.name),
            runtime=self.runtime,
        )
        return func_new

    def eval(self, df):
        ## Check invariant; model inputs must be subset of df columns
        if not set(self.var).issubset(set(df.columns)):
            raise ValueError(
                "Model function `{}` var not a subset of given columns".format(
                    self.name
                )
            )

        ## Predict
        y = self.regressor.predict(df[self.var])
        return DataFrame(data=y, columns=self.out)


## Implementation
# --------------------------------------------------
def _polyridge(
    df, var=None, out=None, n_dim=None, n_degree=None, **kwargs
):
    r"""Low-level interface for polyridge

    Apply the polynomial ridge approximation to seek a low-dimensional subspace
    among inputs that best explain variability in a selected output. Return the
    scikit-style fitted object and modified `var`.

    Args:
        df (DataFrame): Target dataset to reduce
        var (list or None): Variables in df on which to perform dimension reduction.
            Use None to compute with all numeric variables, except `out`.
        out (string): Name of target output
        n_dim (int): Target dimensionality
        n_degree (int): Fitted polynomial (total) order

    Kwargs:
        n_init (int): Number of iterations in optimization

    Returns:
        PolynomialRidgeApproximation: Scikit-style regressor
        list of str: Modified variable list

    Notes:
        - A wrapper for PolynomialRidgeApproximation, originally implemented in the psdr package.

    References:
        - J. M. Hokanson and Paul G. Constantine. Data-driven Polynomial Ridge Approximation Using Variable Projection. SIAM J. Sci. Comput. Vol 40, No 3, pp A1566–A1589, DOI:10.1137/17M1117690.

    """
    ## Check invariants
    if out is None:
        raise ValueError("Must select an `out` column.")
    if not (out in df.columns):
        raise ValueError("out={} is not a column in provided dataframe.".format(out))

    if var is None:
        var = set(df.select_dtypes(include=[npnumber]).columns.values)
        var = list(var.difference({out}))
    else:
        var = list(var).copy()
        diff = set(var).difference(set(df.columns))
        if len(diff) > 0:
            raise ValueError(
                "`var` must be subset of `df.columns`\n" "diff = {}".format(diff)
            )

    if (n_degree == 1) and (n_dim > 1):
        raise ValueError(
            "n_dim > 1 cannot work with n_degree == 1; this would be " +
            "an over-parameterized linear model. Try again with n_degree > 1."
        )

    ## Compute subspace reduction
    pr = PolynomialRidgeApproximation(
        n_degree,
        n_dim,
        **kwargs,
    )
    pr.fit(df[var].values, df[out].values)

    return pr, var

## Interfaces
# --------------------------------------------------
@curry
def tran_polyridge(
    df, var=None, out=None, n_dim=None, n_degree=None, seed=None, **kwargs
):
    r"""Polynomial Ridge Approximation for parameter space reduction

    Apply the polynomial ridge approximation to seek a low-dimensional subspace
    among inputs that best explain variability in a selected output.

    Args:
        df (DataFrame): Target dataset to reduce
        var (list or None): Variables in df on which to perform dimension reduction.
            Use None to compute with all numeric variables, except `out`.
        out (string): Name of target output
        n_dim (int): Target dimensionality
        n_degree (int): Fitted polynomial (total) order

    Kwargs:
        n_init (int): Number of iterations in optimization
        seed (int): Seed for random number generator

    Returns:
        DataFrame: Set of weights defining the dimension-reduced space.

    Notes:
        - A wrapper for PolynomialRidgeApproximation, originally implemented in the psdr package.

    References:
        - J. M. Hokanson and Paul G. Constantine. Data-driven Polynomial Ridge Approximation Using Variable Projection. SIAM J. Sci. Comput. Vol 40, No 3, pp A1566–A1589, DOI:10.1137/17M1117690.

    Examples:
        import grama as gr
        from grama.psdr import tf_polyridge
        DF = gr.Intention()

        ## Set up a dataset to reduce
        df_data = (
            gr.df_make(x=range(10))
            >> gr.tf_outer(gr.df_make(y=range(10)))
            >> gr.tf_outer(gr.df_make(z=range(10)))
            >> gr.tf_mutate(f=DF.x - DF.y)
        )

        ## Use polynomial ridge approximation to derive weights
        df_weights = (
            df_data
            >> tf_polyridge(out="f", n_dim=1, n_degree=1)
        )

        ## Construct a shadow plot
        (
            df_data
            >> gr.tf_inner(df_weights)
            >> gr.ggplot(gr.aes("dot", "f"))
            + gr.geom_point()
        )

    """
    ## Set the seed, if provided
    if not (seed is None):
        npseed(seed)
    ## Run the low-level implementation
    pr, var = _polyridge(df, var=var, out=out, n_dim=n_dim, n_degree=n_degree, **kwargs)

    ## Package the results
    df_res = DataFrame(
        data=pr.U.T,
        columns=var
    )
    return df_res

tf_polyridge = add_pipe(tran_polyridge)

@curry
def fit_polyridge(
    df, var=None, out=None, n_dim=None, n_degree=None, domain=None, density=None, seed=None, **kwargs
):
    r"""Polynomial Ridge Approximation fitting routine

    Fit a polynomial model on a subspace reduction of the given variables. Uses
    the variable projection (L2) algorithm of Hokanson and Constantine (2018).

    Args:
        df (DataFrame): Target dataset to reduce
        var (list or None): Variables in df on which to perform dimension reduction.
            Use None to compute with all numeric variables, except `out`.
        out (string): Name of target output
        domain (gr.Domain or None): Domain for fitted model, optional
        density (gr.Density or None): Density for fitted model, optional
        n_dim (int): Target dimensionality
        n_degree (int): Fitted polynomial (total) order

    Kwargs:
        n_init (int): Number of iterations in optimization
        seed (int): Seed for random number generator

    Returns:
        Model: Polynomial (surrogate) model.

    Notes:
        - A wrapper for PolynomialRidgeApproximation, originally implemented in the psdr package.

    References:
        - J. M. Hokanson and Paul G. Constantine. Data-driven Polynomial Ridge Approximation Using Variable Projection. SIAM J. Sci. Comput. Vol 40, No 3, pp A1566–A1589, DOI:10.1137/17M1117690.

    Examples:
        import grama as gr
        DF = gr.Intention()

        ## Set up a dataset to reduce
        df_data = (
            gr.df_make(x=range(10))
            >> gr.tf_outer(gr.df_make(y=range(10)))
            >> gr.tf_outer(gr.df_make(z=range(10)))
            >> gr.tf_mutate(f=DF.x - DF.y)
        )

        ## Use polynomial ridge approximation to fit a surrogate model
        md_poly = (
            df_data
            >> gr.ft_polyridge(out="f", n_dim=1, n_degree=1)
        )

        ## Fit routine is useful for cross-validation of hyperparameters;
        # compare two different hyperparameter settings
        (
            df_data
            >> gr.tf_kfolds(
                ft=gr.ft_polyridge(out="f", n_dim=1, n_degree=1),
                k=5,
            )
        )

        (
            df_data
            >> gr.tf_kfolds(
                ft=gr.ft_polyridge(out="f", n_dim=2, n_degree=2),
                k=5,
            )
        )

    """
    ## Set the seed, if provided
    if not (seed is None):
        npseed(seed)
    ## Run the low-level implementation
    pr, var = _polyridge(df, var=var, out=out, n_dim=n_dim, n_degree=n_degree, **kwargs)

    ## Package the results
    md_res = Model(
        functions=[FunctionPoly(pr, var, [out], "Polyridge ({})".format(out), None)],
        domain=domain,
        density=density,
    )
    return md_res

ft_polyridge = add_pipe(fit_polyridge)




""" Polynomial functions """
# (c) 2019 Jeffrey M. Hokanson (jeffrey@hokanson.us)

from __future__ import print_function
import numpy as np
import scipy.linalg
from copy import copy
from .basis import *
from .function import BaseFunction

from numpy.polynomial.legendre import legroots

__all__ = ['PolynomialFunction', 'PolynomialApproximation']


def linear_fit(A, b, norm = 2, bound = None):
	r""" solve the linear optimization problem subject to constraints
	"""
	assert norm in [1,2,np.inf], "Invalid norm specified"
	assert bound in [None, 'lower', 'upper'], "invalid bound specified"

	if norm == 2 and bound == None:
		return scipy.linalg.lstsq(A, b)[0]
	else:
                raise NotImplementedError


class PolynomialFunction(BaseFunction):
	r""" A polynomial function in a Legendre basis


	Parameters
	----------
	dimension: int
		Input dimension
	degree: int
		Degree of polynomial
	coef: array-like
		Coefficients of polynomial
	"""
	def __init__(self, basis, coef):
		self.basis = basis
		self.coef = np.array(coef)

	def roots(self):
		return self.basis.roots(self.coef)


	def V(self, X):
		return self.basis.V(X)

	def DV(self, X):
		return self.basis.DV(X)

	def DDV(self, X):
		return self.basis.DDV(X)

	def eval(self, X):
		if len(X.shape) == 1:
			return self.V(X.reshape(1,-1)).dot(self.coef).reshape(1)
		else:
			return self.V(X).dot(self.coef)

	def grad(self, X):
		if len(X.shape) == 1:
			one_d = True
			X = X.reshape(1,-1)
		else:
			one_d = False

		DV = self.DV(X)
		# Compute gradient on projected space
		Df = np.tensordot(DV, self.coef, axes = (1,0))
		# Inflate back to whole space
		if one_d:
			return Df.reshape(X.shape[1])
		else:
			return Df

	def hessian(self, X):
		if len(X.shape) == 1:
			one_d = True
			X = X.reshape(1,-1)
		else:
			one_d = False

		DDV = self.DDV(X)
		DDf = np.tensordot(DDV, self.coef, axes = (1,0))
		if one_d:
			return DDf.reshape(X.shape[1], X.shape[1])
		else:
			return DDf


class PolynomialApproximation(PolynomialFunction):
	r""" Construct a polynomial approximation

	Parameters
	----------
	degree: int
		Degree of polynomial
	basis: ['arnold', 'legendre', 'monomial', 'chebyshev', 'laguerre', 'hermite']
		Basis in which to express the polynomial
	norm: [1, 2, np.inf]
		Norm in which to find the approximation
	bound: [None, 'lower', 'upper']
		If None, construct approximation in the specified norm;
		if 'lower' or 'upper', additionally enforce the constraint that
		the approximation is below or above the measured samples
	"""
	def __init__(self, degree, basis = 'legendre', norm = 2, bound = None):

		degree = int(degree)
		assert degree >= 0, "Degree must be positive"
		self.degree = degree

		assert basis in ['arnoldi', 'legendre', 'monomial', 'chebyshev', 'laguerre', 'hermite']
		self.basis_name = copy(basis)

		self.basis = None

		assert bound in [None, 'lower', 'upper']
		self.bound = bound

		assert norm in [1,2, np.inf]
		self.norm = norm


	def fit(self, X, fX):
		M, m = X.shape

		# Since we don't know the input dimension until we get the data, we initialize the basis here
		if self.basis_name == 'arnoldi':
			self.basis = ArnoldiPolynomialBasis(self.degree, X = X)
		elif self.basis_name == 'legendre':
			self.basis = LegendreTensorBasis(self.degree, X = X)
		elif self.basis_name == 'monomial':
			self.basis = MonomialTensorBasis(self.degree, X = X)
		elif self.basis_name == 'chebyshev':
			self.basis = ChebyshevTensorBasis(self.degree, X = X)
		elif self.basis_name == 'laguerre':
			self.basis = LaguerreTensorBasis(self.degree, X = X)
		elif self.basis_name == 'hermite':
			self.basis = HermiteTensorBasis(self.degree, X = X)

		# Construct Vandermonde matrix
		V = self.basis.V(X)

		self.coef = linear_fit(V, fX, norm = self.norm, bound = self.bound)



__all__ = ["make_channel_nondim", "make_channel"]

from grama import cp_bounds, cp_copula_independence, cp_vec_function, cp_marginals
from grama import Model, df_make, marg_mom
from numpy import exp
from pandas import DataFrame

## Define the PSAAP 1d Channel model; dimensionless form
def make_channel_nondim():
    r"""Make 1d channel model; dimensionless form

    Instantiates a model for particle and fluid temperature rise; particles are suspended in a fluid with bulk velocity along a square cross-section channel. The walls of said channel are transparent, and radiation heats the particles as they travel down the channel.

    References:
        Banko, A.J. "RADIATION ABSORPTION BY INERTIAL PARTICLES IN A TURBULENT SQUARE DUCT FLOW" (2018) PhD Thesis, Stanford University, Chapter 2

    """
    md = (
        Model("1d Particle-laden Channel with Radiation; Dimensionless Form")
        >> cp_vec_function(
            fun=lambda df: df_make(
                beta=120 * (1 + df.Phi_M * df.chi)
            ),
            var=["Phi_M", "chi"],
            out=["beta"],
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                T_f=(df.Phi_M * df.chi) /
                    (1 + df.Phi_M * df.chi) * (
                        df.I * df.xst - df.beta**(-1) * df.I * (1 - exp(-df.beta * df.xst))
                    ),
                T_p=1 /
                    (1 + df.Phi_M * df.chi) * (
                        df.Phi_M * df.chi * df.I * df.xst
                      + df.beta**(-1) * df.I * (1 - exp(-df.beta * df.xst))
                    ),
            ),
            var=["xst", "Phi_M", "chi", "I", "beta"],
            out=["T_f", "T_p"],
        )
        >> cp_bounds(
            ## Dimensionless axial location (-)
            xst=(0, 5),
        )
        >> cp_marginals(
            ## Mass loading ratio (-)
            Phi_M={"dist": "uniform", "loc": 0, "scale": 1},
            ## Particle-fluid heat capacity ratio (-)
            chi={"dist": "uniform", "loc": 0.1, "scale": 0.9},
            ## Normalized radiative intensity (-)
            I={"dist": "uniform", "loc": 0.1, "scale": 0.9},
         )
        >> cp_copula_independence()
    )

    return md

## Define the PSAAP 1d Channel model; dimensional form
def make_channel():
    r"""Make 1d channel model; dimensional form

    Instantiates a model for particle and fluid temperature rise; particles are suspended in a fluid with bulk velocity along a square cross-section channel. The walls of said channel are transparent, and radiation heats the particles as they travel down the channel.

    Note that this takes the same inputs as the builtin dataset `df_channel`.

    References:
        Banko, A.J. "RADIATION ABSORPTION BY INERTIAL PARTICLES IN A TURBULENT SQUARE DUCT FLOW" (2018) PhD Thesis, Stanford University, Chapter 2

    Examples:

    >>> import grama as gr
    >>> from grama.data import df_channel
    >>> from grama.models import make_channel
    >>> md_channel = make_channel()

    >>> (
    >>>     df_channel
    >>>     >> gr.tf_md(md_channel)

    >>>     >> gr.ggplot(gr.aes("T_f", "T_norm"))
    >>>     + gr.geom_abline(slope=1, intercept=0, linetype="dashed")
    >>>     + gr.geom_point()
    >>>     + gr.labs(x="1D Model", y="3D DNS")
    >>> )

    """
    md = (
        Model("1d Particle-laden Channel with Radiation; Dimensional Form")
        >> cp_vec_function(
            fun=lambda df: df_make(
                Re=df.U * df.H / df.nu_f,
                chi=df.cp_p / df.cp_f,
                Pr=df.nu_f / df.alpha_f,
                Phi_M=df.rho_p * 0.524 * df.d_p**3 * df.n / df.rho_f,
                tau_flow=df.L / df.U,
                tau_pt=(df.rho_p * df.cp_p * 0.318 * df.d_p) / df.h_p,
                tau_rad=(df.rho_p * df.cp_p * 0.667 * df.d_p * df.T_0)
                       /(df.Q_abs * 0.78 * df.I_0),
            ),
            var=[
                "U",        # Fluid bulk velocity
                "H",        # Channel width
                "nu_f",     # Fluid kinematic viscosity
                "cp_p",     # Particle isobaric heat capacity
                "cp_f",     # Fluid isobaric heat capacity
                "alpha_f",  # Fluid thermal diffusivity
                "rho_p",    # Particle density
                "rho_f",    # Fluid density
                "d_p",      # Particle diameter
                "n",        # Particle number density
                "h_p",      # Particle-to-gas convection coefficient
                "T_0",      # Initial temperature
                "Q_abs",    # Particle radiation absorption coefficient
                "I_0",      # Incident radiation
            ],
            out=[
                "Re",       # Reynolds number
                "Pr",       # Prandtl number
                "chi",      # Particle-fluid heat capacity ratio
                "Phi_M",    # Mass Loading Ratio
                "tau_flow", # Fluid residence time
                "tau_pt",   # Particle thermal time constant
                "tau_rad",  # Particle temperature doubling time (approximate)
            ],
            name="Dimensionless Numbers",
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                ## Let xi = x / L
                xst=(df.xi * df.L) / df.H / df.Re / df.Pr,
                ## Assume an optically-thin scenario; I/I_0 = 1
                Is=df.Re * df.Pr * (df.H / df.L) * (df.tau_flow / df.tau_rad) * 1,
                beta=df.Re * df.Pr * (df.H / df.L) * (df.tau_flow / df.tau_pt)
                    *(1 + df.Phi_M * df.chi),
            ),
            var=["xi", "chi", "H", "L", "Phi_M", "tau_flow", "tau_rad", "tau_pt"],
            out=[
                "xst",   # Flow-normalized channel axial location
                "Is",    # Normalized heat flux
                "beta",  # Spatial development coefficient
            ],
            name="Intermediate Dimensionless Numbers",
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                T_f=(df.Phi_M * df.chi) /
                    (1 + df.Phi_M * df.chi) * (
                        df.Is * df.xst - df.Is / df.beta * (1 - exp(-df.beta * df.xst))
                    ),
                T_p=1 / (1 + df.Phi_M * df.chi) * (
                        df.Phi_M * df.chi * df.Is * df.xst
                      + df.Is / df.beta * (1 - exp(-df.beta * df.xst))
                    ),
            ),
            var=["xst", "Phi_M", "chi", "Is", "beta"],
            out=["T_f", "T_p"],
        )
        >> cp_bounds(
            ## Normalized axial location; xi = x/L (-)
            xi=(0, 1),
        )
        >> cp_marginals(
            ## Channel width (m)
            H={"dist": "uniform", "loc": 0.038, "scale": 0.004},
            ## Channel length (m)
            L={"dist": "uniform", "loc": 0.152, "scale": 0.016},
            ## Fluid bulk velocity (m/s)
            U={"dist": "uniform", "loc": 1, "scale": 2.5},
            ## Fluid kinematic viscosity (m^2/s)
            nu_f={"dist": "uniform", "loc": 1.4e-5, "scale": 0.1e-5},
            ## Particle isobaric heat capacity (J/(kg K))
            cp_p={"dist": "uniform", "loc": 100, "scale": 900},
            ## Fluid isobaric heat capacity (J/(kg K))
            cp_f={"dist": "uniform", "loc": 1000, "scale": 1000},
            ## Fluid thermal diffusivity (m^2/s)
            alpha_f={"dist": "uniform", "loc": 50e-6, "scale": 50e-6},
            ## Particle density (kg / m^3)
            rho_p={"dist": "uniform", "loc": 1e3, "scale": 9e3},
            ## Fluid density (kg / m^3)
            rho_f={"dist": "uniform", "loc": 0.5, "scale": 1.0},
            ## Particle diameter (m)
            d_p={"dist": "uniform", "loc": 1e-6, "scale": 9e-6},
            ## Particle number density (1 / m^3)
            n={"dist": "uniform", "loc": 9.5e9, "scale": 1.0e9},
            ## Particle-to-gas convection coefficient (W / (m^2 K))
            h_p={"dist": "uniform", "loc": 1e3, "scale": 9e3},
            ## Initial temperature (K)
            T_0={"dist": "uniform", "loc": 285, "scale": 30},
            ## Particle radiation absorption coefficient (-)
            Q_abs={"dist": "uniform", "loc": 0.25, "scale": 0.50},
            ## Incident radiation (W/m^2)
            I_0={"dist": "uniform", "loc": 9.5e6, "scale": 1.0e6},
         )
        >> cp_copula_independence()
    )

    return md



from grama.models import make_trajectory_linear
from grama.data import df_trajectory_full
import grama as gr
import os 

n_process = int(os.cpu_count()/2)

md_trajectory = make_trajectory_linear()

df_fit = (
    md_trajectory
    >> gr.ev_nls(df_data=df_trajectory_full, n_restart=10, n_process=n_process)
)



import numpy as np
import scipy.spatial.distance

def cdist(X1, X2, L = None):
	r""" A convience wrapper around a weighted l2-norm distance calculation
	"""
	if len(X1.shape) == 1:
		X1 = np.reshape(X1,(1,-1))
	if len(X2.shape) == 1:
		X2 = np.reshape(X2,(1,-1))

	if L is not None:
		X1 = (L @ X1.T).T
		X2 = (L @ X2.T).T

	return scipy.spatial.distance.cdist(X1, X2)



import numpy as np
import pandas as pd
from scipy.stats import norm
import unittest

from context import grama as gr
from context import data

##==============================================================================
## grouping test functions
##==============================================================================
class TestGroup(unittest.TestCase):
    def test_group_attributes(self):
        d = data.df_diamonds >> gr.tf_group_by("cut")
        self.assertTrue(hasattr(d, "_grouped_by"))
        self.assertTrue(
            d._grouped_by == ["cut",]
        )
